# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import pytest

from astrapy.exceptions import DataAPIResponseException
from astrapy.info import (
    AlterTableAddColumns,
    AlterTableAddVectorize,
    AlterTableDropColumns,
    AlterTableDropVectorize,
    CreateTableDefinition,
    TableIndexDefinition,
    TableIndexDescriptor,
    TableIndexOptions,
    TableIndexType,
    TableKeyValuedColumnTypeDescriptor,
    TablePrimaryKeyDescriptor,
    TableScalarColumnTypeDescriptor,
    TableTextIndexDefinition,
    TableTextIndexOptions,
    TableUDTColumnDescriptor,
    TableValuedColumnTypeDescriptor,
    TableVectorColumnTypeDescriptor,
    TableVectorIndexDefinition,
    TableVectorIndexOptions,
    VectorServiceOptions,
)

from ..conftest import IS_ASTRA_DB

if TYPE_CHECKING:
    from astrapy import Database


def _remove_apisupport(def_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Strip apiSupport keys from columns since its presence
    is != between sending and receiving.
    """

    def _clean_col(col_dict: dict[str, Any]) -> dict[str, Any]:
        return {k: v for k, v in col_dict.items() if k != "apiSupport"}

    return {
        k: v if k != "columns" else {colk: _clean_col(colv) for colk, colv in v.items()}
        for k, v in def_dict.items()
    }


def _remove_definition(def_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Strip definition keys from columns since its presence
    is != between sending and receiving. Useful for UDT columns
    """

    def _clean_col(col_dict: dict[str, Any]) -> dict[str, Any]:
        # 'definition' may appear nested, e.g. for a map with UDT values
        return {
            k: _clean_col(v) if isinstance(v, dict) else v
            for k, v in col_dict.items()
            if k != "definition"
        }

    return {
        k: v if k != "columns" else {colk: _clean_col(colv) for colk, colv in v.items()}
        for k, v in def_dict.items()
    }


class TestTableLifecycle:
    @pytest.mark.describe("test of create/verify/delete tables, sync")
    def test_table_basic_crd_sync(
        self,
        sync_database: Database,
    ) -> None:
        pre_tables = sync_database.list_table_names()
        created_table_names = {
            "table_whole_obj",
            "table_whole_dict",
            "table_hybrid",
            "table_fluent",
        }
        assert set(pre_tables) & created_table_names == set()

        table_whole_obj_definition = CreateTableDefinition(
            columns={
                "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
                "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
                "p_boolean": TableScalarColumnTypeDescriptor(column_type="boolean"),
                "p_float": TableScalarColumnTypeDescriptor(column_type="float"),
                "p_set": TableValuedColumnTypeDescriptor(
                    column_type="set", value_type="int"
                ),
                "p_map": TableKeyValuedColumnTypeDescriptor(
                    column_type="map", key_type="text", value_type="int"
                ),
                "p_vector": TableVectorColumnTypeDescriptor(
                    column_type="vector", dimension=191, service=None
                ),
                "p_vectorize": TableVectorColumnTypeDescriptor(
                    column_type="vector",
                    dimension=1024,
                    service=VectorServiceOptions(
                        provider="mistral",
                        model_name="mistral-embed",
                    ),
                ),
            },
            primary_key=TablePrimaryKeyDescriptor(
                partition_by=[
                    "p_text",
                    "p_int",
                ],
                partition_sort={"p_boolean": -1, "p_float": 1},
            ),
        )
        sync_database.create_table(
            "table_whole_obj",
            definition=table_whole_obj_definition,
        )
        sync_database.create_table(
            "table_whole_dict",
            definition={
                "columns": {
                    "p_text": {"type": "text"},
                    "p_int": {"type": "int"},
                    "p_boolean": {"type": "boolean"},
                    "p_float": {"type": "float"},
                    "p_set": {"type": "set", "valueType": "int"},
                    "p_map": {"type": "map", "keyType": "text", "valueType": "int"},
                    "p_vector": {"type": "vector", "dimension": 191},
                    "p_vectorize": {
                        "type": "vector",
                        "dimension": 1024,
                        "service": {
                            "provider": "mistral",
                            "modelName": "mistral-embed",
                        },
                    },
                },
                "primaryKey": {
                    "partitionBy": [
                        "p_text",
                        "p_int",
                    ],
                    "partitionSort": {"p_boolean": -1, "p_float": 1},
                },
            },
        )
        sync_database.create_table(
            "table_hybrid",
            definition={
                "columns": {
                    "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
                    "p_int": {"type": "int"},
                    "p_boolean": TableScalarColumnTypeDescriptor(column_type="boolean"),
                    "p_float": {"type": "float"},
                    "p_set": TableValuedColumnTypeDescriptor(
                        column_type="set", value_type="int"
                    ),
                    "p_map": {"type": "map", "keyType": "text", "valueType": "int"},
                    "p_vector": TableVectorColumnTypeDescriptor(
                        column_type="vector", dimension=191, service=None
                    ),
                    "p_vectorize": {
                        "type": "vector",
                        "dimension": 1024,
                        "service": {
                            "provider": "mistral",
                            "modelName": "mistral-embed",
                        },
                    },
                },
                "primaryKey": TablePrimaryKeyDescriptor(
                    partition_by=[
                        "p_text",
                        "p_int",
                    ],
                    partition_sort={"p_boolean": -1, "p_float": 1},
                ),
            },
        )

        ct_fluent_definition = (
            CreateTableDefinition.builder()
            .add_column("p_text", "text")
            .add_column("p_int", "int")
            .add_column("p_boolean", "boolean")
            .add_scalar_column("p_float", "float")
            .add_set_column("p_set", "int")
            .add_map_column("p_map", "text", "int")
            .add_vector_column("p_vector", dimension=191)
            .add_vector_column(
                "p_vectorize",
                dimension=1024,
                service=VectorServiceOptions(
                    provider="mistral",
                    model_name="mistral-embed",
                ),
            )
            .add_partition_by(["p_text", "p_int"])
            .add_partition_sort({"p_boolean": -1, "p_float": 1})
            .build()
        )
        table_fluent = sync_database.create_table(
            "table_fluent",
            definition=ct_fluent_definition,
        )

        # definition and info
        # Compare classes, not dicts, due to representational ambiguity:
        assert (
            CreateTableDefinition._from_dict(
                _remove_apisupport(table_fluent.definition().as_dict())
            )
            == table_whole_obj_definition
        )
        if IS_ASTRA_DB:
            fl_info = table_fluent.info()
            assert fl_info.name == "table_fluent"

        with pytest.raises(DataAPIResponseException):
            sync_database.create_table(
                "table_fluent",
                definition=ct_fluent_definition,
            )
        sync_database.create_table(
            "table_fluent",
            definition=ct_fluent_definition,
            if_not_exists=True,
        )

        assert (
            set(sync_database.list_table_names()) - set(pre_tables)
            == created_table_names
        )

        created_table_descs = [
            table_desc
            for table_desc in sync_database.list_tables()
            if table_desc.name in created_table_names
        ]

        for table_desc in created_table_descs[1:]:
            assert table_desc.definition == created_table_descs[0].definition

        for created_table_name in created_table_names:
            sync_database.drop_table(created_table_name)

        assert sync_database.list_table_names() == pre_tables

    @pytest.mark.describe("test of table create/delete index, sync")
    def test_tableindex_basic_crd_sync(
        self,
        sync_database: Database,
    ) -> None:
        table = sync_database.create_table(
            "table_for_indexes",
            definition=CreateTableDefinition(
                columns={
                    "p_key": TableScalarColumnTypeDescriptor(column_type="text"),
                    "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
                    "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
                    "p_vector": TableVectorColumnTypeDescriptor(
                        column_type="vector", dimension=191, service=None
                    ),
                    "p_vector_sm": TableVectorColumnTypeDescriptor(
                        column_type="vector", dimension=37, service=None
                    ),
                },
                primary_key=TablePrimaryKeyDescriptor(
                    partition_by=["p_key"],
                    partition_sort={},
                ),
            ),
        )

        table.create_index(
            "tfi_idx_p_text",
            "p_text",
            options=TableIndexOptions(
                ascii=False,
                normalize=True,
                case_sensitive=False,
            ),
        )
        table.create_index(
            "tfi_idx_p_int",
            "p_int",
        )
        table.create_vector_index(
            "tfi_idx_p_vector_sm",
            "p_vector_sm",
            options=TableVectorIndexOptions(
                metric="cosine",
            ),
        )
        table.create_vector_index(
            "tfi_idx_p_vector",
            "p_vector",
            options=TableVectorIndexOptions(
                source_model="openai-v3-large",
            ),
        )

        expected_indexes = {
            "tfi_idx_p_text",
            "tfi_idx_p_int",
            "tfi_idx_p_vector",
            "tfi_idx_p_vector_sm",
        }
        assert set(table.list_index_names()) == expected_indexes
        # this data needs metric and source_model always specified
        # to match how the API behaves. Also metric is auto-selected based
        # on sourceModel for example.
        expected_index_list = [
            TableIndexDescriptor(
                name="tfi_idx_p_text",
                definition=TableIndexDefinition(
                    column="p_text",
                    options=TableIndexOptions(
                        ascii=False,
                        normalize=True,
                        case_sensitive=False,
                    ),
                ),
            ),
            TableIndexDescriptor(
                name="tfi_idx_p_int",
                definition=TableIndexDefinition(
                    column="p_int",
                    options=TableIndexOptions(),
                ),
            ),
            TableIndexDescriptor(
                name="tfi_idx_p_vector_sm",
                definition=TableVectorIndexDefinition(
                    column="p_vector_sm",
                    options=TableVectorIndexOptions(
                        metric="cosine",
                        source_model="other",
                    ),
                ),
            ),
            TableIndexDescriptor(
                name="tfi_idx_p_vector",
                definition=TableVectorIndexDefinition(
                    column="p_vector",
                    options=TableVectorIndexOptions(
                        metric="dot_product",
                        source_model="openai-v3-large",
                    ),
                ),
            ),
        ]
        assert sorted(table.list_indexes(), key=lambda id: id.name) == sorted(
            expected_index_list, key=lambda id: id.name
        )

        sync_database.drop_table_index("tfi_idx_p_text")
        sync_database.drop_table_index("tfi_idx_p_int")
        sync_database.drop_table_index("tfi_idx_p_vector")
        sync_database.drop_table_index("tfi_idx_p_vector_sm")
        sync_database.drop_table(table.name)

    @pytest.mark.describe("test of alter table, sync")
    def test_alter_table_sync(
        self,
        sync_database: Database,
    ) -> None:
        orig_table_def = CreateTableDefinition(
            columns={
                "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
                "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
                "p_vector": TableVectorColumnTypeDescriptor(
                    column_type="vector", dimension=1024, service=None
                ),
            },
            primary_key=TablePrimaryKeyDescriptor(
                partition_by=["p_text"],
                partition_sort={},
            ),
        )
        table = sync_database.create_table(
            "table_to_alter",
            definition=orig_table_def,
        )

        try:
            table.alter(
                operation=AlterTableAddColumns.coerce({"columns": {"added_1": "float"}})
            )
            table.alter(
                # one with a dictionary to parse
                operation={"drop": {"columns": ["added_1"]}}
            )
            table.alter(
                operation=AlterTableAddVectorize.coerce(
                    {
                        "columns": {
                            "p_vector": {
                                "provider": "mistral",
                                "modelName": "mistral-embed",
                            }
                        }
                    }
                )
            )
            table.alter(
                operation=AlterTableDropVectorize.coerce({"columns": ["p_vector"]})
            )
            # back to the original table:
            # Compare classes, not dicts, due to representational ambiguity:
            assert (
                CreateTableDefinition._from_dict(
                    _remove_apisupport(table.definition().as_dict())
                )
                == orig_table_def
            )
        finally:
            table.drop()

    @pytest.mark.describe("test of collection indexes, sync")
    def test_table_collectionindexes_sync(
        self,
        sync_database: Database,
    ) -> None:
        table_colidx_def = CreateTableDefinition(
            columns={
                "id": TableScalarColumnTypeDescriptor(column_type="text"),
                "t": TableScalarColumnTypeDescriptor(column_type="text"),
                "set_int": TableValuedColumnTypeDescriptor(
                    column_type="set",
                    value_type="int",
                ),
                "set_int2": TableValuedColumnTypeDescriptor(
                    column_type="set",
                    value_type="int",
                ),
                "list_int": TableValuedColumnTypeDescriptor(
                    column_type="list",
                    value_type="int",
                ),
                "list_int2": TableValuedColumnTypeDescriptor(
                    column_type="list",
                    value_type="int",
                ),
                "map_text_int_e": TableKeyValuedColumnTypeDescriptor(
                    key_type="text",
                    value_type="int",
                ),
                "map_text_int_k": TableKeyValuedColumnTypeDescriptor(
                    key_type="text",
                    value_type="int",
                ),
                "map_text_int_v": TableKeyValuedColumnTypeDescriptor(
                    key_type="text",
                    value_type="int",
                ),
            },
            primary_key=TablePrimaryKeyDescriptor(
                partition_by=["id"],
                partition_sort={},
            ),
        )
        table = sync_database.create_table(
            "table_collectionindexes",
            definition=table_colidx_def,
        )

        try:
            # create, list and drop baseline + various collection-column indexes
            idx_t_t_column = "t"
            idx_t_t_options = {
                "caseSensitive": False,
                "normalize": False,
                "ascii": True,
            }
            idx_t_set_int_column = "set_int"
            idx_t_set_int_column2 = {"set_int2": "$values"}
            idx_t_list_int_column = "list_int"
            idx_t_list_int_column2 = {"list_int2": "$values"}
            idx_t_map_text_int_e_column = "map_text_int_e"
            idx_t_map_text_int_k_column = {"map_text_int_k": "$keys"}
            idx_t_map_text_int_v_column = {"map_text_int_v": "$values"}

            table.create_index("idx_t_t", idx_t_t_column, options=idx_t_t_options)
            table.create_index("idx_t_set_int", idx_t_set_int_column)
            table.create_index("idx_t_set_int2", idx_t_set_int_column2)
            table.create_index("idx_t_list_int", idx_t_list_int_column)
            table.create_index("idx_t_list_int2", idx_t_list_int_column2)
            table.create_index("idx_t_map_text_int_e", idx_t_map_text_int_e_column)
            table.create_index("idx_t_map_text_int_k", idx_t_map_text_int_k_column)
            table.create_index("idx_t_map_text_int_v", idx_t_map_text_int_v_column)

            listed_indexes = sorted(
                table.list_indexes(),
                key=lambda idx_desc: idx_desc.name,
            )
            expected_indexes = sorted(
                [
                    TableIndexDescriptor(
                        name="idx_t_t",
                        definition=TableIndexDefinition(
                            column=idx_t_t_column,
                            options=TableIndexOptions.coerce(idx_t_t_options),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_set_int",
                        definition=TableIndexDefinition(
                            column={idx_t_set_int_column: "$values"},
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_set_int2",
                        definition=TableIndexDefinition(
                            column=idx_t_set_int_column2,
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_list_int",
                        definition=TableIndexDefinition(
                            column={idx_t_list_int_column: "$values"},
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_list_int2",
                        definition=TableIndexDefinition(
                            column=idx_t_list_int_column2,
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_map_text_int_e",
                        definition=TableIndexDefinition(
                            column=idx_t_map_text_int_e_column,
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_map_text_int_k",
                        definition=TableIndexDefinition(
                            column=idx_t_map_text_int_k_column,
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                    TableIndexDescriptor(
                        name="idx_t_map_text_int_v",
                        definition=TableIndexDefinition(
                            column=idx_t_map_text_int_v_column,
                            options=TableIndexOptions(),
                        ),
                        index_type=TableIndexType.REGULAR,
                    ),
                ],
                key=lambda idx_desc: idx_desc.name,
            )
            assert listed_indexes == expected_indexes

            table.database.drop_table_index("idx_t_t")
            table.database.drop_table_index("idx_t_set_int")
            table.database.drop_table_index("idx_t_set_int2")
            table.database.drop_table_index("idx_t_list_int")
            table.database.drop_table_index("idx_t_list_int2")
            table.database.drop_table_index("idx_t_map_text_int_e")
            table.database.drop_table_index("idx_t_map_text_int_k")
            table.database.drop_table_index("idx_t_map_text_int_v")
        finally:
            table.drop()

    # TODO: open to Astra DB once available
    @pytest.mark.skipif(
        IS_ASTRA_DB,
        reason="Text indexes not on Astra DB yet",
    )
    @pytest.mark.describe("test of text indexes, sync")
    def test_table_textindexes_sync(
        self,
        sync_database: Database,
    ) -> None:
        table_textidx_def = CreateTableDefinition(
            columns={
                "id": TableScalarColumnTypeDescriptor(column_type="text"),
                "txt_d": TableScalarColumnTypeDescriptor(column_type="text"),
                "txt_s": TableScalarColumnTypeDescriptor(column_type="text"),
                "txt_l": TableScalarColumnTypeDescriptor(column_type="text"),
            },
            primary_key=TablePrimaryKeyDescriptor(
                partition_by=["id"],
                partition_sort={},
            ),
        )
        table = sync_database.create_table(
            "table_textindexes",
            definition=table_textidx_def,
        )

        try:
            # create, list and drop various analyzer text-column indexes
            tx_id_opts_s = TableTextIndexOptions(analyzer="whitespace")
            tx_id_opts_l = TableTextIndexOptions(
                analyzer={
                    "tokenizer": {"name": "standard", "args": {}},
                    "filters": [
                        {"name": "lowercase"},
                        {"name": "stop"},
                        {"name": "porterstem"},
                        {"name": "asciifolding"},
                    ],
                    "charFilters": [],
                },
            )
            table.create_text_index("idx_txt_d", "txt_d")
            table.create_text_index("idx_txt_s", "txt_s", options=tx_id_opts_s)
            table.create_text_index("idx_txt_l", "txt_l", options=tx_id_opts_l)

            listed_indexes = sorted(
                table.list_indexes(),
                key=lambda idx_desc: idx_desc.name,
            )
            expected_indexes = sorted(
                [
                    TableIndexDescriptor(
                        name="idx_txt_d",
                        definition=TableTextIndexDefinition(
                            column="txt_d",
                            options=TableTextIndexOptions(analyzer="standard"),
                        ),
                        index_type=TableIndexType.TEXT,
                    ),
                    TableIndexDescriptor(
                        name="idx_txt_s",
                        definition=TableTextIndexDefinition(
                            column="txt_s",
                            options=tx_id_opts_s,
                        ),
                        index_type=TableIndexType.TEXT,
                    ),
                    TableIndexDescriptor(
                        name="idx_txt_l",
                        definition=TableTextIndexDefinition(
                            column="txt_l",
                            options=tx_id_opts_l,
                        ),
                        index_type=TableIndexType.TEXT,
                    ),
                ],
                key=lambda idx_desc: idx_desc.name,
            )
            assert listed_indexes == expected_indexes

            table.database.drop_table_index("idx_txt_d")
            table.database.drop_table_index("idx_txt_s")
            table.database.drop_table_index("idx_txt_l")
        finally:
            table.drop()

    @pytest.mark.skipif(
        "ASTRAPY_TEST_UDT" not in os.environ,
        reason="UDT testing not enabled",
    )
    @pytest.mark.describe("test of create/verify/delete table with a simple UDT, sync")
    def test_table_simpleudt_crd_sync(
        self,
        sync_database: Database,
        simple_udt: str,
    ) -> None:
        udt_col_desc = TableUDTColumnDescriptor(udt_name=simple_udt)
        table_simple_udt_def = CreateTableDefinition(
            columns={
                "id": TableScalarColumnTypeDescriptor("text"),
                "udt_map": TableKeyValuedColumnTypeDescriptor(
                    key_type=TableScalarColumnTypeDescriptor("ascii"),
                    value_type=udt_col_desc,
                ),
                "udt_set": TableValuedColumnTypeDescriptor(
                    column_type="set",
                    value_type=udt_col_desc,
                ),
                "udt_list": TableValuedColumnTypeDescriptor(
                    column_type="list",
                    value_type=udt_col_desc,
                ),
                "udt_scalar": udt_col_desc,
            },
            primary_key=TablePrimaryKeyDescriptor(
                partition_by=["id"],
                partition_sort={},
            ),
        )
        try:
            table = sync_database.create_table(
                "table_simple_udt",
                definition=table_simple_udt_def,
            )
            # Compare classes, not dicts, due to representational ambiguity:
            assert (
                CreateTableDefinition._from_dict(
                    _remove_definition(_remove_apisupport(table.definition().as_dict()))
                )
                == table_simple_udt_def
            )

            add_udt_columns = AlterTableAddColumns(
                columns={
                    "altered_udt_map": TableKeyValuedColumnTypeDescriptor(
                        key_type=TableScalarColumnTypeDescriptor("ascii"),
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_set": TableValuedColumnTypeDescriptor(
                        column_type="set",
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_list": TableValuedColumnTypeDescriptor(
                        column_type="list",
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_scalar": udt_col_desc,
                },
            )
            table.alter(add_udt_columns)
            drop_udt_columns = AlterTableDropColumns(
                columns=["udt_map", "udt_set", "udt_list", "udt_scalar"]
            )
            table.alter(drop_udt_columns)
            altered_table_simple_udt_def = CreateTableDefinition(
                columns={
                    "id": TableScalarColumnTypeDescriptor("text"),
                    "altered_udt_map": TableKeyValuedColumnTypeDescriptor(
                        key_type=TableScalarColumnTypeDescriptor("ascii"),
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_set": TableValuedColumnTypeDescriptor(
                        column_type="set",
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_list": TableValuedColumnTypeDescriptor(
                        column_type="list",
                        value_type=udt_col_desc,
                    ),
                    "altered_udt_scalar": udt_col_desc,
                },
                primary_key=TablePrimaryKeyDescriptor(
                    partition_by=["id"],
                    partition_sort={},
                ),
            )
            # Compare classes, not dicts, due to representational ambiguity:
            assert (
                CreateTableDefinition._from_dict(
                    _remove_definition(_remove_apisupport(table.definition().as_dict()))
                )
                == altered_table_simple_udt_def
            )
        finally:
            table.drop()

    @pytest.mark.describe("test of table index creation dict call patterns, sync")
    def test_tableindex_creationdictpatterns_sync(
        self,
        sync_database: Database,
    ) -> None:
        try:
            table = sync_database.create_table(
                "table_for_idxcallpattern_dict",
                definition=CreateTableDefinition(
                    columns={
                        "id": TableScalarColumnTypeDescriptor("text"),
                        "a1": TableScalarColumnTypeDescriptor("ascii"),
                        "a2": TableScalarColumnTypeDescriptor("ascii"),
                        "a3": TableScalarColumnTypeDescriptor("ascii"),
                        "s1": TableScalarColumnTypeDescriptor("text"),
                        "s2": TableScalarColumnTypeDescriptor("text"),
                        "s3": TableScalarColumnTypeDescriptor("text"),
                        "v1": TableVectorColumnTypeDescriptor(dimension=2),
                        "v2": TableVectorColumnTypeDescriptor(dimension=2),
                        "v3": TableVectorColumnTypeDescriptor(dimension=2),
                    },
                    primary_key=TablePrimaryKeyDescriptor(
                        partition_by=["id"],
                        partition_sort={},
                    ),
                ),
            )

            # create_index, working patterns
            table.create_index("test_idx_a1", "a1", options={"ascii": True})
            table.create_index("test_idx_a2", "a2")
            table.create_index(
                "test_idx_a3", definition={"column": "a3", "options": {"ascii": True}}
            )
            # create_index, failing patterns
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_index("x", "x", definition={}, options={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_index("x", "x", definition={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_index("x", definition={}, options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_index("x", options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_index("x")

            # create_vector_index, working patterns
            table.create_vector_index(
                "test_idx_v1", "v1", options={"metric": "EUCLIDEAN"}
            )
            table.create_vector_index("test_idx_v2", "v2")
            table.create_vector_index(
                "test_idx_v3",
                definition={"column": "v3", "options": {"metric": "EUCLIDEAN"}},
            )
            # create_vector_index, failing patterns
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_vector_index("x", "x", definition={}, options={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_vector_index("x", "x", definition={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_vector_index("x", definition={}, options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_vector_index("x", options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_vector_index("x")

            # create_text_index, working patterns
            table.create_text_index(
                "test_idx_s1", "s1", options={"analyzer": "whitespace"}
            )
            table.create_text_index("test_idx_s2", "s2")
            table.create_text_index(
                "test_idx_s3",
                definition={"column": "s3", "options": {"analyzer": "whitespace"}},
            )
            # create_text_index, failing patterns
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_text_index("x", "x", definition={}, options={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_text_index("x", "x", definition={})
            with pytest.raises(ValueError, match="cannot be used simultaneously"):
                table.create_text_index("x", definition={}, options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_text_index("x", options={})
            with pytest.raises(ValueError, match="is required for an index"):
                table.create_text_index("x")

            # some indexes should formally coincide:
            indexes = table.list_indexes()
            options_per_idx = {
                idx.name: idx.definition.as_dict().get("options") for idx in indexes
            }
            assert options_per_idx["test_idx_a1"] == options_per_idx["test_idx_a3"]
            assert options_per_idx["test_idx_v1"] == options_per_idx["test_idx_v3"]
            assert options_per_idx["test_idx_s1"] == options_per_idx["test_idx_s3"]
        finally:
            sync_database.drop_table("table_for_idxcallpattern_dict")

    @pytest.mark.describe("test of table index creation object call patterns, sync")
    def test_tableindex_creationobjpatterns_sync(
        self,
        sync_database: Database,
    ) -> None:
        try:
            table = sync_database.create_table(
                "table_for_idxcallpattern_obj",
                definition=CreateTableDefinition(
                    columns={
                        "id": TableScalarColumnTypeDescriptor("text"),
                        "a1": TableScalarColumnTypeDescriptor("ascii"),
                        "a2": TableScalarColumnTypeDescriptor("ascii"),
                        "a3": TableScalarColumnTypeDescriptor("ascii"),
                        "s1": TableScalarColumnTypeDescriptor("text"),
                        "s2": TableScalarColumnTypeDescriptor("text"),
                        "s3": TableScalarColumnTypeDescriptor("text"),
                        "v1": TableVectorColumnTypeDescriptor(dimension=2),
                        "v2": TableVectorColumnTypeDescriptor(dimension=2),
                        "v3": TableVectorColumnTypeDescriptor(dimension=2),
                    },
                    primary_key=TablePrimaryKeyDescriptor(
                        partition_by=["id"],
                        partition_sort={},
                    ),
                ),
            )

            # create_index, working patterns
            table.create_index(
                "test_idx_a1", "a1", options=TableIndexOptions(ascii=True)
            )
            table.create_index("test_idx_a2", "a2")
            table.create_index(
                "test_idx_a3",
                definition=TableIndexDefinition(
                    column="a3", options=TableIndexOptions(ascii=True)
                ),
            )

            # create_vector_index, working patterns
            table.create_vector_index(
                "test_idx_v1", "v1", options=TableVectorIndexOptions(metric="EUCLIDEAN")
            )
            table.create_vector_index("test_idx_v2", "v2")
            table.create_vector_index(
                "test_idx_v3",
                definition=TableVectorIndexDefinition(
                    column="v3", options=TableVectorIndexOptions(metric="EUCLIDEAN")
                ),
            )

            # create_text_index, working patterns
            table.create_text_index(
                "test_idx_s1",
                "s1",
                options=TableTextIndexOptions(analyzer="whitespace"),
            )
            table.create_text_index("test_idx_s2", "s2")
            table.create_text_index(
                "test_idx_s3",
                definition=TableTextIndexDefinition(
                    column="s3", options=TableTextIndexOptions(analyzer="whitespace")
                ),
            )

            # some indexes should formally coincide:
            indexes = table.list_indexes()
            options_per_idx = {
                idx.name: idx.definition.as_dict().get("options") for idx in indexes
            }
            assert options_per_idx["test_idx_a1"] == options_per_idx["test_idx_a3"]
            assert options_per_idx["test_idx_v1"] == options_per_idx["test_idx_v3"]
            assert options_per_idx["test_idx_s1"] == options_per_idx["test_idx_s3"]
        finally:
            sync_database.drop_table("table_for_idxcallpattern_obj")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: F401

from __future__ import annotations

import pytest


@pytest.mark.describe("test namespace")
def test_namespace() -> None:
    import astrapy

    assert str(astrapy.admin) != ""
    assert str(astrapy.api_options) != ""
    assert str(astrapy.authentication) != ""
    assert str(astrapy.client) != ""
    assert str(astrapy.collection) != ""
    assert str(astrapy.constants) != ""
    assert str(astrapy.data) != ""
    assert str(astrapy.database) != ""
    assert str(astrapy.exceptions) != ""
    assert str(astrapy.ids) != ""
    assert str(astrapy.info) != ""
    assert str(astrapy.results) != ""
    assert str(astrapy.settings) != ""
    assert str(astrapy.utils) != ""

    assert str(astrapy.admin.AstraDBAdmin) != ""
    assert str(astrapy.api_options.APIOptions) != ""
    assert str(astrapy.authentication.TokenProvider) != ""
    assert str(astrapy.client.DataAPIClient) != ""
    assert str(astrapy.collection.Collection) != ""
    assert str(astrapy.constants.VectorMetric.DOT_PRODUCT) != ""
    assert str(astrapy.data.collection) != ""
    assert str(astrapy.database.Database) != ""
    assert str(astrapy.exceptions.DevOpsAPIException) != ""
    assert str(astrapy.ids.uuid6) != ""
    assert str(astrapy.info.AstraDBDatabaseInfo) != ""
    assert str(astrapy.results.CollectionDeleteResult) != ""
    assert str(astrapy.settings.defaults) != ""
    assert str(astrapy.utils.request_tools) != ""


@pytest.mark.describe("test imports")
def test_imports() -> None:
    from astrapy import (
        AstraDBAdmin,
        AstraDBDatabaseAdmin,
        AsyncCollection,
        AsyncDatabase,
        AsyncTable,
        Collection,
        DataAPIClient,
        DataAPIDatabaseAdmin,
        Database,
        Table,
    )
    from astrapy.admin import (
        ParsedAPIEndpoint,
        parse_api_endpoint,
    )
    from astrapy.api_options import (
        APIOptions,
        DataAPIURLOptions,
        DevOpsAPIURLOptions,
        SerdesOptions,
        TimeoutOptions,
    )
    from astrapy.authentication import (
        AWSEmbeddingHeadersProvider,
        EmbeddingAPIKeyHeaderProvider,
        EmbeddingHeadersProvider,
        StaticTokenProvider,
        UsernamePasswordTokenProvider,
    )
    from astrapy.constants import (
        DefaultIdType,
        Environment,
        MapEncodingMode,
        ReturnDocument,
        SortMode,
        VectorMetric,
    )
    from astrapy.cursors import (
        AbstractCursor,
        AsyncCollectionFindAndRerankCursor,
        AsyncCollectionFindCursor,
        AsyncTableFindCursor,
        CollectionFindAndRerankCursor,
        CollectionFindCursor,
        CursorState,
        RerankedResult,
        TableFindCursor,
    )
    from astrapy.data_types import (
        DataAPIDate,
        DataAPIDictUDT,
        DataAPIDuration,
        DataAPIMap,
        DataAPISet,
        DataAPITime,
        DataAPITimestamp,
        DataAPIVector,
    )
    from astrapy.exceptions import (
        CollectionDeleteManyException,
        CollectionInsertManyException,
        CollectionUpdateManyException,
        CursorException,
        DataAPIErrorDescriptor,
        DataAPIException,
        DataAPIHttpException,
        DataAPIResponseException,
        DataAPITimeoutException,
        DevOpsAPIErrorDescriptor,
        DevOpsAPIException,
        DevOpsAPIHttpException,
        DevOpsAPIResponseException,
        DevOpsAPITimeoutException,
        MultiCallTimeoutManager,
        TableInsertManyException,
        TooManyDocumentsToCountException,
        TooManyRowsToCountException,
        UnexpectedDataAPIResponseException,
        UnexpectedDevOpsAPIResponseException,
    )
    from astrapy.ids import (
        UUID,
        ObjectId,
        uuid1,
        uuid3,
        uuid4,
        uuid5,
        uuid6,
        uuid7,
        uuid8,
    )
    from astrapy.info import (
        AlterTableAddColumns,
        AlterTableAddVectorize,
        AlterTableDropColumns,
        AlterTableDropVectorize,
        AlterTypeAddFields,
        AlterTypeOperation,
        AlterTypeRenameFields,
        AstraDBAdminDatabaseInfo,
        AstraDBDatabaseInfo,
        CollectionDefaultIDOptions,
        CollectionDefinition,
        CollectionDescriptor,
        CollectionInfo,
        CollectionLexicalOptions,
        CollectionRerankOptions,
        CollectionVectorOptions,
        ColumnType,
        CreateTableDefinition,
        CreateTypeDefinition,
        EmbeddingProvider,
        EmbeddingProviderAuthentication,
        EmbeddingProviderModel,
        EmbeddingProviderParameter,
        EmbeddingProviderToken,
        FindEmbeddingProvidersResult,
        FindRerankingProvidersResult,
        ListTableDefinition,
        ListTableDescriptor,
        ListTypeDescriptor,
        RerankingProvider,
        RerankingProviderAuthentication,
        RerankingProviderModel,
        RerankingProviderParameter,
        RerankingProviderToken,
        RerankServiceOptions,
        TableAPIIndexSupportDescriptor,
        TableAPISupportDescriptor,
        TableBaseIndexDefinition,
        TableIndexDefinition,
        TableIndexDescriptor,
        TableIndexOptions,
        TableInfo,
        TableKeyValuedColumnType,
        TableKeyValuedColumnTypeDescriptor,
        TablePrimaryKeyDescriptor,
        TableScalarColumnTypeDescriptor,
        TableTextIndexDefinition,
        TableTextIndexOptions,
        TableUDTColumnDescriptor,
        TableUnsupportedColumnTypeDescriptor,
        TableUnsupportedIndexDefinition,
        TableValuedColumnType,
        TableValuedColumnTypeDescriptor,
        TableVectorColumnTypeDescriptor,
        TableVectorIndexDefinition,
        TableVectorIndexOptions,
        VectorServiceOptions,
    )
    from astrapy.results import (
        CollectionDeleteResult,
        CollectionInsertManyResult,
        CollectionInsertOneResult,
        CollectionUpdateResult,
        OperationResult,
        TableInsertManyResult,
        TableInsertOneResult,
    )
    from astrapy.utils.api_options import defaultAPIOptions
    from astrapy.utils.document_paths import (
        escape_field_names,
        unescape_field_path,
    )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar

from astrapy.data.info.table_descriptor.table_columns import (
    TableColumnTypeDescriptor,
)
from astrapy.utils.parsing import _warn_residual_keys

AYO = TypeVar("AYO", bound="AlterTypeOperation")


@dataclass
class AlterTypeOperation(ABC):
    """
    An abstract class representing a generic "alter type" operation, i.e. a change
    to be applied to a user-defined type (UDT) stored on the database. Concrete
    implementations are used to represent operations such as adding/renaming fields.

    `AlterTypeOperation` objects are used in the Database's `alter_type` method.

    Please consult the documentation of the concrete subclasses for more info.
    """

    _name: str

    @abstractmethod
    def as_dict(self) -> dict[str, Any]: ...

    @staticmethod
    def from_full_dict(operation_dict: dict[str, Any]) -> AlterTypeOperation:
        """
        Inspect a provided dictionary and make it into the correct concrete subclass
        of AlterTypeOperation depending on its contents.

        Note: while the nature of the operation must be the top-level single key of
        the (nested) dictionary parameter to this method (such as "add" or
        "rename"), the resulting `AlterTypeOperation` object encodes the content
        of the corresponding value. Likewise, calling the `as_dict()` method of
        the result from this method does not return the whole original input, rather
        the "one level in" part (see the example provided here).

        Args:
            operation_dict: a dictionary such as `{"add": ...}`, whose outermost *value*
            corresponds to the desired operation.

        Returns:
            an `AlterTypeOperation` object chosen after inspection of
                the provided input.

        Example:
            >>> full_dict = {"add": {"fields": {
            ...     "fld1": "text", "fld2": {"type": "int"}
            ... }}}
            >>> alter_op = AlterTypeOperation.from_full_dict(full_dict)
            >>> alter_op
            AlterTypeAddFields(fields=[fld1,fld2])
            >>> alter_op.as_dict()
            {'fields': {'fld1': {'type': 'text'}, 'fld2': {'type': 'int'}}}
        """

        key_set = set(operation_dict.keys())
        if key_set == {"add"}:
            return AlterTypeAddFields.coerce(operation_dict["add"])
        elif key_set == {"rename"}:
            return AlterTypeRenameFields.coerce(operation_dict["rename"])
        else:
            raise ValueError(
                f"Cannot parse a dict with keys {', '.join(sorted(key_set))} "
                "into an AlterTypeOperation"
            )

    @classmethod
    @abstractmethod
    def coerce(cls: type[AYO], raw_input: AYO | dict[str, Any]) -> AYO: ...

    @classmethod
    @abstractmethod
    def stack(cls: type[AYO], operations: list[AYO]) -> AYO: ...

    @classmethod
    def stack_by_name(
        cls, operations: list[AlterTypeOperation]
    ) -> dict[str, AlterTypeOperation]:
        grouped_ops: dict[str, list[AlterTypeOperation]] = {}
        for op in operations:
            grouped_ops[op._name] = grouped_ops.get(op._name, []) + [op]

        stacked_op_map: dict[str, AlterTypeOperation] = {
            op_name: op_list[0].stack(op_list)
            for op_name, op_list in grouped_ops.items()
        }
        return stacked_op_map


@dataclass
class AlterTypeAddFields(AlterTypeOperation):
    """
    An object representing the alter-type operation of adding field(s),
    for use in the argument to the database's `alter_type()` method.

    Attributes:
        fields: a mapping between the names of the fields to add and
            `TableColumnTypeDescriptor` objects, formatted in the same way as
            the `columns` attribute of `CreateTableDefinition`.
    """

    fields: dict[str, TableColumnTypeDescriptor]

    def __init__(
        self,
        *,
        fields: dict[str, TableColumnTypeDescriptor | dict[str, Any] | str],
    ) -> None:
        self._name = "add"
        self.fields = {
            fld_n: TableColumnTypeDescriptor.coerce(fld_v)
            for fld_n, fld_v in fields.items()
        }

    def __repr__(self) -> str:
        _fld_desc = f"fields=[{','.join(sorted(self.fields.keys()))}]"
        return f"{self.__class__.__name__}({_fld_desc})"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            "fields": {fld_n: fld_v.as_dict() for fld_n, fld_v in self.fields.items()}
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> AlterTypeAddFields:
        """
        Create an instance of AlterTypeAddFields from a dictionary
        such as one suitable as (partial) command payload.
        """

        _warn_residual_keys(cls, raw_dict, {"fields"})
        return AlterTypeAddFields(fields=raw_dict["fields"])

    @classmethod
    def coerce(
        cls, raw_input: AlterTypeAddFields | dict[str, Any]
    ) -> AlterTypeAddFields:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into an AlterTypeAddFields.
        """

        if isinstance(raw_input, AlterTypeAddFields):
            return raw_input
        else:
            return cls._from_dict(raw_input)

    @classmethod
    def stack(cls, operations: list[AlterTypeAddFields]) -> AlterTypeAddFields:
        fields_dict: dict[str, TableColumnTypeDescriptor | dict[str, Any] | str] = {}
        for ataf in operations:
            fields_dict = {**fields_dict, **ataf.fields}
        return AlterTypeAddFields(fields=fields_dict)


@dataclass
class AlterTypeRenameFields(AlterTypeOperation):
    """
    An object representing the alter-type operation of renaming field(s),
    for use in the argument to the database's `alter_type()` method.

    Attributes:
        fields: a mapping from current to new names for the fields to rename.
    """

    fields: dict[str, str]

    def __init__(self, *, fields: dict[str, str]) -> None:
        self._name = "rename"
        self.fields = fields

    def __repr__(self) -> str:
        _renames = "; ".join(
            [f"'{old_n}' -> '{new_n}'" for old_n, new_n in sorted(self.fields.items())]
        )
        return f"{self.__class__.__name__}(fields=[{_renames}])"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""
        return {
            "fields": self.fields,
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> AlterTypeRenameFields:
        """
        Create an instance of AlterTypeRenameFields from a dictionary
        such as one suitable as (partial) command payload.
        """

        _warn_residual_keys(cls, raw_dict, {"fields"})
        return AlterTypeRenameFields(
            fields={old_n: new_n for old_n, new_n in raw_dict["fields"].items()},
        )

    @classmethod
    def coerce(
        cls, raw_input: AlterTypeRenameFields | dict[str, Any]
    ) -> AlterTypeRenameFields:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into an AlterTypeRenameFields.
        """

        if isinstance(raw_input, AlterTypeRenameFields):
            return raw_input
        else:
            return cls._from_dict(raw_input)

    @classmethod
    def stack(cls, operations: list[AlterTypeRenameFields]) -> AlterTypeRenameFields:
        fields_dict: dict[str, str] = {}
        for ataf in operations:
            fields_dict = {**fields_dict, **ataf.fields}
        return AlterTypeRenameFields(fields=fields_dict)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.info import FindEmbeddingProvidersResult

RESPONSE_DICT_0 = {
    "embeddingProviders": {
        "provider": {
            "displayName": "TheProvider",
            "url": "https://api.provider.com/v1/",
            "supportedAuthentication": {
                "SHARED_SECRET": {
                    "tokens": [
                        {
                            "accepted": "providerKey",
                            "forwarded": "Authorization",
                        },
                    ],
                    "enabled": True,
                },
                "HEADER": {
                    "tokens": [
                        {
                            "accepted": "x-embedding-api-key",
                            "forwarded": "Authorization",
                        },
                    ],
                    "enabled": True,
                },
                "NONE": {
                    "tokens": [],
                    "enabled": True,
                },
            },
            "parameters": [
                {
                    "help": "parameter help",
                    "defaultValue": "x",
                    "displayName": "The Parameter",
                    "hint": "Provide a parameter",
                    "name": "the_parameter",
                    "type": "STRING",
                    "validation": {},
                    "required": False,
                },
            ],
            "models": [
                {
                    "name": "the-model",
                    "vectorDimension": None,
                    "parameters": [
                        {
                            "name": "vectorDimension",
                            "type": "NUMBER",
                            "required": True,
                            "defaultValue": "3072",
                            "validation": {
                                "numericRange": [
                                    256,
                                    3072,
                                ],
                            },
                            "help": "Help for vector dimension",
                        },
                    ],
                },
            ],
        },
    },
}

RESPONSE_DICT_1 = {
    "embeddingProviders": {
        "provider": {
            "displayName": "TheProvider",
            "url": "https://api.provider.com/v1/",
            "supportedAuthentication": {
                "SHARED_SECRET": {
                    "tokens": [
                        {
                            "accepted": "providerKey",
                            "forwarded": "Authorization",
                        },
                    ],
                    "enabled": True,
                },
                "HEADER": {
                    "tokens": [
                        {
                            "accepted": "x-embedding-api-key",
                            "forwarded": "Authorization",
                        },
                    ],
                    "enabled": True,
                },
                "NONE": {
                    "tokens": [],
                    "enabled": True,
                },
            },
            "parameters": [
                {
                    "help": "parameter help",
                    "defaultValue": "x",
                    "displayName": "The Parameter",
                    "hint": "Provide a parameter",
                    "name": "the_parameter",
                    "type": "STRING",
                    "validation": {},
                    "required": False,
                },
            ],
            "models": [
                {
                    "name": "provider/<model>",
                    "apiModelSupport": {
                        "status": "SUPPORTED",
                    },
                    "vectorDimension": None,
                    "parameters": [
                        {
                            "name": "vectorDimension",
                            "type": "NUMBER",
                            "required": True,
                            "defaultValue": "3072",
                            "validation": {
                                "numericRange": [
                                    256,
                                    3072,
                                ],
                            },
                            "help": "Help for vector dimension",
                        },
                    ],
                },
            ],
        },
    },
}


class TestFindEmbeddingProvidersResult:
    @pytest.mark.describe("test of FindEmbeddingProvidersResult parsing and back")
    def test_embedding_providers_result_parsing_and_back_base(self) -> None:
        parsed = FindEmbeddingProvidersResult._from_dict(RESPONSE_DICT_0)
        providers = parsed.embedding_providers
        assert len(providers) == 1
        assert providers["provider"].display_name is not None

        models = providers["provider"].models
        assert len(models) == 1

        dumped = parsed.as_dict()
        # clean out apiModelSupport from generated dict before checking
        for pro_v in dumped["embeddingProviders"].values():
            for mod_v in pro_v["models"]:
                del mod_v["apiModelSupport"]
        assert dumped == RESPONSE_DICT_0

    @pytest.mark.describe(
        "test of FindEmbeddingProvidersResult parsing and back (with apiModelSupport)"
    )
    def test_embedding_providers_result_parsing_and_back_rich(self) -> None:
        parsed = FindEmbeddingProvidersResult._from_dict(RESPONSE_DICT_1)
        providers = parsed.embedding_providers
        assert len(providers) == 1
        assert providers["provider"].display_name is not None

        models = providers["provider"].models
        assert len(models) == 1

        dumped = parsed.as_dict()
        assert dumped == RESPONSE_DICT_1



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.info import VectorServiceOptions


class TestVectorServiceOptions:
    @pytest.mark.describe("test of VectorServiceOptions conversions, base")
    def test_VectorServiceOptions_conversions_base(self) -> None:
        dict0 = {
            "provider": "PRO",
            "modelName": "MOD",
        }
        from_dict0 = VectorServiceOptions._from_dict(dict0)
        assert from_dict0 == VectorServiceOptions(
            provider="PRO",
            model_name="MOD",
        )
        assert from_dict0.as_dict() == dict0

    @pytest.mark.describe("test of VectorServiceOptions conversions, with auth")
    def test_VectorServiceOptions_conversions_auth(self) -> None:
        dict_a = {
            "provider": "PRO",
            "modelName": "MOD",
            "authentication": {
                "type": ["A_T"],
                "field": "value",
            },
        }
        from_dict_a = VectorServiceOptions._from_dict(dict_a)
        assert from_dict_a == VectorServiceOptions(
            provider="PRO",
            model_name="MOD",
            authentication={
                "type": ["A_T"],
                "field": "value",
            },
        )
        assert from_dict_a.as_dict() == dict_a

    @pytest.mark.describe("test of VectorServiceOptions conversions, with params")
    def test_VectorServiceOptions_conversions_params(self) -> None:
        dict_p = {
            "provider": "PRO",
            "modelName": "MOD",
            "parameters": {
                "field1": "value1",
                "field2": 12.3,
                "field3": 123,
                "field4": True,
                "field5": None,
                "field6": {"a": 1},
            },
        }
        from_dict_p = VectorServiceOptions._from_dict(dict_p)
        assert from_dict_p == VectorServiceOptions(
            provider="PRO",
            model_name="MOD",
            parameters={
                "field1": "value1",
                "field2": 12.3,
                "field3": 123,
                "field4": True,
                "field5": None,
                "field6": {"a": 1},
            },
        )
        assert from_dict_p.as_dict() == dict_p

    @pytest.mark.describe(
        "test of VectorServiceOptions conversions, with params and auth"
    )
    def test_VectorServiceOptions_conversions_params_auth(self) -> None:
        dict_ap = {
            "provider": "PRO",
            "modelName": "MOD",
            "authentication": {
                "type": ["A_T"],
                "field": "value",
            },
            "parameters": {
                "field1": "value1",
                "field2": 12.3,
                "field3": 123,
                "field4": True,
                "field5": None,
                "field6": {"a": 1},
            },
        }
        from_dict_ap = VectorServiceOptions._from_dict(dict_ap)
        assert from_dict_ap == VectorServiceOptions(
            provider="PRO",
            model_name="MOD",
            authentication={
                "type": ["A_T"],
                "field": "value",
            },
            parameters={
                "field1": "value1",
                "field2": 12.3,
                "field3": 123,
                "field4": True,
                "field5": None,
                "field6": {"a": 1},
            },
        )
        assert from_dict_ap.as_dict() == dict_ap



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from decimal import Decimal

from astrapy.data_types import DataAPIMap, DataAPISet

_LD = "1.229999999999999982236431605997495353221893310546875"

BASELINE_OBJ = {
    "f": 1.23,
    "fn": float("NaN"),
    "fp": float("Infinity"),
    "fm": float("-Infinity"),
    "i": 123,
    "i0": 0,
    "t": "T",
}
BASELINE_KEY_STR = '[1.23,"Nan","Infinity","-Infinity",123,0,"T"]'
BASELINE_COLUMNS = {
    "f": {"type": "float"},
    "fn": {"type": "float"},
    "fp": {"type": "float"},
    "fm": {"type": "float"},
    "i": {"type": "int"},
    "i0": {"type": "int"},
    "t": {"type": "text"},
}
WDECS_OBJ = {
    "f": 1.23,
    "fn": float("NaN"),
    "fp": float("Infinity"),
    "fm": float("-Infinity"),
    "i": 123,
    "i0": 0,
    "t": "T",
    "d": Decimal("1.23"),
    "df": Decimal(1.23),
    "dl": Decimal(_LD),
    "dn": Decimal("NaN"),
    "dp": Decimal("Infinity"),
    "dm": Decimal("-Infinity"),
}
WDECS_KEY_STR = f'[1.23,"Nan","Infinity","-Infinity",123,0,"T",1.23,{_LD},{_LD},"Nan","Infinity","-Infinity"]'
WDECS_OBJ_COLUMNS = {
    "f": {"type": "float"},
    "fn": {"type": "float"},
    "fp": {"type": "float"},
    "fm": {"type": "float"},
    "i": {"type": "int"},
    "i0": {"type": "int"},
    "t": {"type": "text"},
    "d": {"type": "decimal"},
    "df": {"type": "decimal"},
    "dl": {"type": "decimal"},
    "dn": {"type": "decimal"},
    "dp": {"type": "decimal"},
    "dm": {"type": "decimal"},
}

COLLTYPES_CUSTOM_OBJ_SCHEMA_TRIPLES = [
    (
        False,
        {
            "f": {
                "f": 1.23,
                "fn": float("NaN"),
                "fp": float("Infinity"),
                "fm": float("-Infinity"),
            },
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "float",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPIMap(
                {
                    "f": 1.23,
                    "fn": float("NaN"),
                    "fp": float("Infinity"),
                    "fm": float("-Infinity"),
                }
            ),
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "float",
            },
        },
    ),
    (
        False,
        {
            "f": [
                1.23,
                float("NaN"),
                float("Infinity"),
                float("-Infinity"),
            ],
        },
        {
            "f": {
                "type": "list",
                "valueType": "float",
            },
        },
    ),
    (
        False,
        {
            "f": {
                1.23,
                float("NaN"),
                float("Infinity"),
                float("-Infinity"),
            },
        },
        {
            "f": {
                "type": "set",
                "valueType": "float",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPISet(
                {
                    1.23,
                    float("NaN"),
                    float("Infinity"),
                    float("-Infinity"),
                }
            ),
        },
        {
            "f": {
                "type": "set",
                "valueType": "float",
            },
        },
    ),
    (
        False,
        {
            "f": {
                "i0": 0,
                "im": -19,
                "ip": 384,
            },
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "int",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPIMap(
                {
                    "i0": 0,
                    "im": -19,
                    "ip": 384,
                }
            ),
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "int",
            },
        },
    ),
    (
        False,
        {
            "f": [
                0,
                -19,
                384,
            ],
        },
        {
            "f": {
                "type": "list",
                "valueType": "int",
            },
        },
    ),
    (
        False,
        {
            "f": {
                0,
                -19,
                384,
            },
        },
        {
            "f": {
                "type": "set",
                "valueType": "int",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPISet(
                {
                    0,
                    -19,
                    384,
                }
            ),
        },
        {
            "f": {
                "type": "set",
                "valueType": "int",
            },
        },
    ),
    (
        False,
        {
            "f": {
                "d": Decimal("1.23"),
                "df": Decimal(1.23),
                "dl": Decimal(_LD),
                "dn": Decimal("NaN"),
                "dp": Decimal("Infinity"),
                "dm": Decimal("-Infinity"),
            },
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "decimal",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPIMap(
                {
                    "d": Decimal("1.23"),
                    "df": Decimal(1.23),
                    "dl": Decimal(_LD),
                    "dn": Decimal("NaN"),
                    "dp": Decimal("Infinity"),
                    "dm": Decimal("-Infinity"),
                }
            ),
        },
        {
            "f": {
                "type": "map",
                "keyType": "text",
                "valueType": "decimal",
            },
        },
    ),
    (
        False,
        {
            "f": [
                Decimal("1.23"),
                Decimal(1.23),
                Decimal(_LD),
                Decimal("NaN"),
                Decimal("Infinity"),
                Decimal("-Infinity"),
            ],
        },
        {
            "f": {
                "type": "list",
                "valueType": "decimal",
            },
        },
    ),
    (
        False,
        {
            "f": {
                Decimal("1.23"),
                Decimal(1.23),
                Decimal(_LD),
                Decimal("NaN"),
                Decimal("Infinity"),
                Decimal("-Infinity"),
            },
        },
        {
            "f": {
                "type": "set",
                "valueType": "decimal",
            },
        },
    ),
    (
        True,
        {
            "f": DataAPISet(
                {
                    Decimal("1.23"),
                    Decimal(1.23),
                    Decimal(_LD),
                    Decimal("NaN"),
                    Decimal("Infinity"),
                    Decimal("-Infinity"),
                }
            ),
        },
        {
            "f": {
                "type": "set",
                "valueType": "decimal",
            },
        },
    ),
]
COLLTYPES_TRIPLE_IDS = [
    "map/float",
    "DataAPIMap/float",
    "list/float",
    "set/float",
    "DataAPISet/float",
    "map/int",
    "DataAPIMap/int",
    "list/int",
    "set/int",
    "DataAPISet/int",
    "map/Decimal",
    "DataAPIMap/Decimal",
    "list/Decimal",
    "set/Decimal",
    "DataAPISet/Decimal",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the ObjectIds and UUIDn conversions, 'idiomatic' imports
"""

from __future__ import annotations

import json

import pytest

from astrapy.data.utils.collection_converters import (
    postprocess_collection_response,
    preprocess_collection_payload,
)
from astrapy.ids import UUID, ObjectId
from astrapy.utils.api_options import defaultSerdesOptions


@pytest.mark.describe("test of serdes for ids")
def test_ids_serdes() -> None:
    f_u1 = UUID("8ccd6ff8-e61b-11ee-a2fc-7df4a8c4164b")  # uuid1
    f_u3 = UUID("6fa459ea-ee8a-3ca4-894e-db77e160355e")  # uuid3
    f_u4 = UUID("4f16cba8-1115-43ab-aa39-3a9c29f37db5")  # uuid4
    f_u5 = UUID("886313e1-3b8a-5372-9b90-0c9aee199e5d")  # uuid5
    f_u6 = UUID("1eee61b9-8f2d-69ad-8ebb-5054d2a1a2c0")  # uuid6
    f_u7 = UUID("018e57e5-f586-7ed6-be55-6b0de3041116")  # uuid7
    f_u8 = UUID("018e57e5-fbcd-8bd4-b794-be914f2c4c85")  # uuid8
    f_oi = ObjectId("65f9cfa0d7fabb3f255c25a1")

    full_structure = {
        "f_u1": f_u1,
        "f_u3": f_u3,
        "f_u4": f_u4,
        "f_u5": f_u5,
        "f_u6": f_u6,
        "f_u7": f_u7,
        "f_u8": f_u8,
        "f_oi": f_oi,
    }

    normalized = preprocess_collection_payload(
        full_structure,
        options=defaultSerdesOptions,
    )
    json.dumps(normalized)
    assert normalized is not None
    restored = postprocess_collection_response(
        normalized,
        options=defaultSerdesOptions,
    )
    assert restored == full_structure



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import pickle

import pytest

from astrapy.data_types import DataAPITimestamp
from astrapy.utils.date_utils import (
    AVERAGE_YEAR_MS,
    EPOCH_YEAR,
    _year_to_unix_timestamp_ms,
    _year_to_unix_timestamp_ms_backward,
    _year_to_unix_timestamp_ms_forward,
)


class TestDataAPITimestamp:
    @pytest.mark.describe("test of DataAPITimestamp class")
    def test_dataapitimestamp_lifecycle(self) -> None:
        # in-range use (far enough from valid range for tz-dependent conversions)
        Y1_1_1_MILLIS = -62135596800000
        Y10K_12_31_MILLIS = 253402300800000
        ONE_DAY_AND_EPSILON_MILLIS = 86400000 + 555
        TEN_YEARS_MILLIS = 10 * 365 * 86400000
        ts_y00001 = DataAPITimestamp(Y1_1_1_MILLIS + ONE_DAY_AND_EPSILON_MILLIS)
        ts_y00001_2 = DataAPITimestamp(Y1_1_1_MILLIS + ONE_DAY_AND_EPSILON_MILLIS)
        ts_y10000 = DataAPITimestamp(Y10K_12_31_MILLIS - ONE_DAY_AND_EPSILON_MILLIS)

        assert ts_y00001 == ts_y00001_2
        assert ts_y00001 != ts_y10000
        assert (
            DataAPITimestamp.from_datetime(ts_y00001.to_datetime(tz=None)) == ts_y00001
        )
        assert (
            DataAPITimestamp.from_datetime(ts_y10000.to_datetime(tz=None)) == ts_y10000
        )

        # (not: DataAPITimestamp is not supposed to respect sub-millisecond precision.)
        dt1 = datetime.datetime(
            2024, 10, 27, 11, 22, 33, 831000, tzinfo=datetime.timezone.utc
        )
        assert dt1 == DataAPITimestamp.from_datetime(dt1).to_datetime(
            tz=datetime.timezone.utc
        )

        # out-of-ranges
        ts_y00009_bc = DataAPITimestamp(Y1_1_1_MILLIS - TEN_YEARS_MILLIS)
        ts_y10010_ad = DataAPITimestamp(Y10K_12_31_MILLIS + TEN_YEARS_MILLIS)
        with pytest.raises(ValueError, match="is out of range"):
            ts_y00009_bc.to_datetime(tz=None)
        with pytest.raises(ValueError, match="is out of range"):
            ts_y10010_ad.to_datetime(tz=None)

        # algebra
        assert ts_y00001 < ts_y10000
        assert ts_y00001 <= ts_y10000
        assert ts_y00001 <= ts_y00001_2
        assert ts_y10000 >= ts_y00001
        assert ts_y10000 >= ts_y00001
        assert ts_y00001 >= ts_y00001_2
        assert ts_y00001 - ts_y00001 == 0
        assert ts_y10000 - ts_y00001 > 0
        assert ts_y00001 - ts_y10000 < 0

    @pytest.mark.describe("test of DataAPITimestamp year-to-epoch utilities")
    def test_dataapitimestamp_year_timestamp(self) -> None:
        assert _year_to_unix_timestamp_ms_forward(1970) == 0
        assert _year_to_unix_timestamp_ms_backward(1970) == 0
        assert _year_to_unix_timestamp_ms(1970) == 0
        for y in range(1, 10000, 15):
            py_y = datetime.datetime(y, 1, 1, 0, 0, 0).replace(
                tzinfo=datetime.timezone.utc
            )
            py_ts_s = py_y.timestamp()
            util_ts_ms = _year_to_unix_timestamp_ms(y)
            assert py_ts_s * 1000 == util_ts_ms

    @pytest.mark.describe("test of DataAPITimestamp parsing, failures")
    def test_dataapitimestamp_parsing_failures(self) -> None:
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("boo!")

        # signs for various lengths of the 'year digits' portion
        # valid cases
        DataAPITimestamp.from_string("0000-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("-2024-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("+112024-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("+000024-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("-112024-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("0124-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("-0124-10-29T01:25:37.123Z")
        # incorrect cases for sign of year
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("+0000-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("-0000-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("+124-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("+2024-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("+-2024-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("-+2024-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("124-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("+124-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("-124-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("123456-10-29T01:25:37.123Z")

        DataAPITimestamp.from_string("2024-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("2024-10-29T01:25:37Z")
        DataAPITimestamp.from_string("2024-10-29T01:25:37.123+01:23")
        DataAPITimestamp.from_string("2024-10-29T01:25:37.123-23:12")
        DataAPITimestamp.from_string("2024-10-29T01:25:37+01:23")
        DataAPITimestamp.from_string("2024-10-29T01:25:37-23:12")
        DataAPITimestamp.from_string("-123456-10-29T01:25:37.123Z")
        DataAPITimestamp.from_string("2024-2-3T4:5:6Z")
        DataAPITimestamp.from_string("2024-02-03T04:05:06Z")

        DataAPITimestamp.from_string("1972-02-29T04:05:06Z")
        DataAPITimestamp.from_string("1972-02-29T04:05:06+23:59")
        DataAPITimestamp.from_string("1972-02-29T04:05:06-23:59")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("1971-02-29T04:05:06Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("1971-02-29T04:05:06+23:59")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("1971-02-29T04:05:06-23:59")

        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("a2024-10-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-q29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T01:25:37.123Zb")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-13-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-32T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T25:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T01:61:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T01:25:61.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T01:25:37.123+25:43")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-29T01:25:37.123-27:91")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-00-29T01:25:37.123Z")
        with pytest.raises(ValueError):
            DataAPITimestamp.from_string("2024-10-00T01:25:37.123Z")

    @pytest.mark.describe(
        "test of DataAPITimestamp sign-of-year parse and stringify consistency"
    )
    def test_dataapitimestamp_sign_of_year_consistency(self) -> None:
        def parse_and_back_year_str(year_str: str) -> None:
            # just testing that years come out as they get in with numdigits, sign
            ts_string = f"{year_str}-01-12T12:00:00.000Z"
            ts = DataAPITimestamp.from_string(ts_string)
            ts_string_1 = ts.to_string()
            assert ts_string == ts_string_1

        parse_and_back_year_str("2024")
        parse_and_back_year_str("0000")
        parse_and_back_year_str("-2024")
        parse_and_back_year_str("+112024")
        parse_and_back_year_str("-112024")
        parse_and_back_year_str("0124")
        parse_and_back_year_str("-0124")
        parse_and_back_year_str("-123456")
        parse_and_back_year_str("1972")

    @pytest.mark.describe("test of DataAPITimestamp parsing, consistency")
    def test_dataapitimestamp_parsing_consistency(self) -> None:
        for year in range(2, 9999, 50):
            for offset_hour in [-20, -2, 0, 2, 20]:
                for offset_minute in [0, 25, 45]:
                    for millisecond in [0, 789]:
                        py_dt_0 = datetime.datetime(
                            year, 4, 27, 12, 34, 56, millisecond * 1000
                        ).replace(tzinfo=datetime.timezone.utc)
                        py_dt = py_dt_0 - datetime.timedelta(
                            hours=offset_hour,
                            minutes=offset_minute,
                        )
                        _fraction = "" if millisecond == 0 else f".{millisecond:03}"
                        _offset = (
                            "Z"
                            if offset_hour == 0 and offset_minute == 0
                            else f"{offset_hour:+03}:{offset_minute:02}"
                        )
                        ts_string = f"{year:04}-04-27T12:34:56{_fraction}{_offset}"
                        py_ts_ms = int(py_dt.timestamp() * 1000)
                        util_ts_ms = DataAPITimestamp.from_string(
                            ts_string
                        ).timestamp_ms
                        assert py_ts_ms == util_ts_ms

    @pytest.mark.describe(
        "test of DataAPITimestamp, back to tuple and string, with datetimes"
    )
    def test_dataapitimestamp_back_to_tuple_string_wdatetimes(self) -> None:
        """
        Test of various string/tuple-powered conversions and their consistency.
        In range covered by 'datetime'.

        Plan:
                                                    --> datetime[4]
                                                   /
        datetime[0] --> timestamp --> DataAPITS[1] ---> tuple[2]
                                                   \
                                                    --> string --> DataAPITS[3]
                
        followed by checks that:
            [0] == [2]
            [1] == [3]
            [0] == [4]
        """

        for test_y in [
            1,
            8,
            100,
            1604,
            1969,
            1970,
            1971,
            1972,
            2024,
            2100,
            2777,
            5432,
            7776,
            9999,
        ]:
            for test_mo in [1, 2, 7, 10, 12]:
                for test_h in [0, 1, 17, 23]:
                    for test_m in [0, 59]:
                        for test_s in [0, 59]:
                            for test_us in [0, 654000, 999000]:
                                test_dt = datetime.datetime(
                                    test_y,
                                    test_mo,
                                    28,
                                    test_h,
                                    test_m,
                                    test_s,
                                    test_us,
                                    tzinfo=datetime.timezone.utc,
                                )
                                dapi_ts = DataAPITimestamp.from_datetime(test_dt)
                                dapi_tuple = dapi_ts.timetuple()
                                # [0] == [2]
                                assert test_dt.timetuple()[:6] == dapi_tuple[:6]
                                # allow for 1 ms fluctuations (from rounding and the like):
                                assert (
                                    abs(test_dt.microsecond // 1000 - dapi_tuple[6])
                                    <= 1
                                )

                                # [1] == [3]
                                dapi_str = dapi_ts.to_string()
                                assert dapi_ts == DataAPITimestamp.from_string(dapi_str)

                                # [0] == [4]
                                # allow for 1 ms fluctuations (from rounding and the like):
                                assert (
                                    abs(
                                        (
                                            dapi_ts.to_datetime(
                                                tz=datetime.timezone.utc
                                            )
                                            - test_dt
                                        ).total_seconds()
                                    )
                                    <= 0.001
                                )

    @pytest.mark.describe(
        "test of DataAPITimestamp, back to tuple and string, no datetimes"
    )
    def test_dataapitimestamp_back_to_tuple_string_nodatetimes(self) -> None:
        """
        Test of various string/tuple-powered conversions and their consistency.

        In range NOT covered by 'datetime':
        spanning a wider year range, loop by timestamp

            timestamp -> DataAPITS[0] -> strin -> DataAPITS[1]

        checking that [0] == [1] between 50000 BC and 50000 AD.
        """

        min_ts = AVERAGE_YEAR_MS * (-50000 - EPOCH_YEAR)
        max_ts = AVERAGE_YEAR_MS * (50000 - EPOCH_YEAR)
        num_steps = 200
        step_ms = (max_ts - min_ts) // num_steps
        for test_ms in range(min_ts, max_ts, step_ms):
            dapi_ts = DataAPITimestamp(test_ms)
            dapi_ts2 = DataAPITimestamp.from_string(dapi_ts.to_string())
            assert dapi_ts == dapi_ts2

    @pytest.mark.describe("test pickling of DataAPITimestamp")
    def test_dataapitimestamp_pickle(self) -> None:
        the_timestamp = DataAPITimestamp(1234567890)
        assert pickle.loads(pickle.dumps(the_timestamp)) == the_timestamp



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time

import pytest

from astrapy import DataAPIClient, Database
from astrapy.constants import DefaultIdType, VectorMetric
from astrapy.ids import UUID, ObjectId
from astrapy.info import (
    AstraDBDatabaseInfo,
    CollectionDefaultIDOptions,
    CollectionDefinition,
    CollectionDescriptor,
    CollectionLexicalOptions,
    CollectionRerankOptions,
    CollectionVectorOptions,
    RerankServiceOptions,
)

from ..conftest import (
    IS_ASTRA_DB,
    SECONDARY_KEYSPACE,
    TEST_COLLECTION_NAME,
    DataAPICredentials,
    DataAPICredentialsInfo,
    DefaultCollection,
)


class TestCollectionDDLSync:
    @pytest.mark.describe("test of collection creation, get, and then drop, sync")
    def test_collection_lifecycle_sync(
        self,
        sync_database: Database,
    ) -> None:
        TEST_LOCAL_COLLECTION_NAME = "test_local_coll"
        TEST_LOCAL_COLLECTION_NAME_B = "test_local_coll_b"
        col1 = sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME,
            definition=CollectionDefinition(
                vector=CollectionVectorOptions(
                    dimension=123,
                    metric=VectorMetric.EUCLIDEAN,
                ),
                indexing={"deny": ["a", "b", "c"]},
            ),
        )
        # test other creation methods (no-op since just created)
        col1_dict = sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME,
            definition={
                "vector": {
                    "dimension": 123,
                    "metric": VectorMetric.EUCLIDEAN,
                },
                "indexing": {"deny": ["a", "b", "c"]},
            },
        )
        assert col1_dict == col1
        fluent_definition1 = (
            CollectionDefinition.builder()
            .set_vector_dimension(123)
            .set_vector_metric(VectorMetric.EUCLIDEAN)
            .set_indexing("deny", ["a", "b", "c"])
            .build()
        )
        col1_fluent = sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME,
            definition=fluent_definition1,
        )
        assert col1_fluent == col1

        sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME_B,
            definition=CollectionDefinition(
                indexing={"allow": ["z"]},
            ),
        )
        lc_response = list(sync_database.list_collections())
        #
        rerank_lexical_portion = {
            "rerank": {
                "enabled": True,
                "service": {
                    "provider": "nvidia",
                    "modelName": "nvidia/llama-3.2-nv-rerankqa-1b-v2",
                },
            },
            "lexical": {
                "enabled": True,
                "analyzer": "standard",
            },
        }
        expected_coll_descriptor = CollectionDescriptor._from_dict(
            {
                "name": TEST_LOCAL_COLLECTION_NAME,
                "options": {
                    "vector": {
                        "dimension": 123,
                        "metric": "euclidean",
                        "sourceModel": "other",
                    },
                    "indexing": {"deny": ["a", "b", "c"]},
                    **rerank_lexical_portion,
                },
            },
        )
        expected_coll_descriptor_b = CollectionDescriptor._from_dict(
            {
                "name": TEST_LOCAL_COLLECTION_NAME_B,
                "options": {
                    "indexing": {"allow": ["z"]},
                    **rerank_lexical_portion,
                },
            },
        )
        assert expected_coll_descriptor in lc_response
        assert expected_coll_descriptor_b in lc_response
        #
        col2 = sync_database.get_collection(TEST_LOCAL_COLLECTION_NAME)
        assert col1 == col2
        sync_database.drop_collection(TEST_LOCAL_COLLECTION_NAME)
        sync_database.drop_collection(TEST_LOCAL_COLLECTION_NAME)
        sync_database.drop_collection(TEST_LOCAL_COLLECTION_NAME_B)

    @pytest.mark.describe("test of default_id_type in creating collections, sync")
    def test_collection_default_id_type_sync(
        self,
        sync_database: Database,
    ) -> None:
        ID_TEST_COLLECTION_NAME_ROOT = "id_type_test_"

        col = sync_database.create_collection(
            ID_TEST_COLLECTION_NAME_ROOT + DefaultIdType.UUID,
            definition=CollectionDefinition(
                default_id=CollectionDefaultIDOptions(
                    default_id_type=DefaultIdType.UUID,
                ),
            ),
        )
        col_options = col.options()
        assert col_options is not None
        assert col_options.default_id is not None
        assert col_options.default_id.default_id_type == DefaultIdType.UUID
        i1res = col.insert_one({"role": "probe"})
        assert isinstance(i1res.inserted_id, UUID)
        doc = col.find_one({})
        assert doc is not None
        assert isinstance(doc["_id"], UUID)
        col.drop()

        time.sleep(2)
        col = sync_database.create_collection(
            ID_TEST_COLLECTION_NAME_ROOT + DefaultIdType.UUIDV6,
            definition=CollectionDefinition(
                default_id=CollectionDefaultIDOptions(
                    default_id_type=DefaultIdType.UUIDV6,
                ),
            ),
        )
        col_options = col.options()
        assert col_options is not None
        assert col_options.default_id is not None
        assert col_options.default_id.default_id_type == DefaultIdType.UUIDV6
        i1res = col.insert_one({"role": "probe"})
        assert isinstance(i1res.inserted_id, UUID)
        assert i1res.inserted_id.version == 6
        doc = col.find_one({})
        assert doc is not None
        assert isinstance(doc["_id"], UUID)
        assert doc["_id"].version == 6
        col.drop()

        time.sleep(2)
        col = sync_database.create_collection(
            ID_TEST_COLLECTION_NAME_ROOT + DefaultIdType.UUIDV7,
            definition=CollectionDefinition(
                default_id=CollectionDefaultIDOptions(
                    default_id_type=DefaultIdType.UUIDV7,
                ),
            ),
        )
        col_options = col.options()
        assert col_options is not None
        assert col_options.default_id is not None
        assert col_options.default_id.default_id_type == DefaultIdType.UUIDV7
        i1res = col.insert_one({"role": "probe"})
        assert isinstance(i1res.inserted_id, UUID)
        assert i1res.inserted_id.version == 7
        doc = col.find_one({})
        assert doc is not None
        assert isinstance(doc["_id"], UUID)
        assert doc["_id"].version == 7
        col.drop()

        time.sleep(2)
        col = sync_database.create_collection(
            ID_TEST_COLLECTION_NAME_ROOT + DefaultIdType.DEFAULT,
            definition=CollectionDefinition(
                default_id=CollectionDefaultIDOptions(
                    default_id_type=DefaultIdType.DEFAULT,
                ),
            ),
        )
        col_options = col.options()
        assert col_options is not None
        assert col_options.default_id is not None
        assert col_options.default_id.default_id_type == DefaultIdType.DEFAULT
        col.drop()

        time.sleep(2)
        col = sync_database.create_collection(
            ID_TEST_COLLECTION_NAME_ROOT + DefaultIdType.OBJECTID,
            definition=CollectionDefinition(
                default_id=CollectionDefaultIDOptions(
                    default_id_type=DefaultIdType.OBJECTID,
                ),
            ),
        )
        col_options = col.options()
        assert col_options is not None
        assert col_options.default_id is not None
        assert col_options.default_id.default_id_type == DefaultIdType.OBJECTID
        i1res = col.insert_one({"role": "probe"})
        assert isinstance(i1res.inserted_id, ObjectId)
        doc = col.find_one({})
        assert doc is not None
        assert isinstance(doc["_id"], ObjectId)
        col.drop()

    @pytest.mark.describe("test of collection drop, sync")
    def test_collection_drop_sync(self, sync_database: Database) -> None:
        col = sync_database.create_collection(
            name="sync_collection_to_drop",
            definition=CollectionDefinition(
                vector=CollectionVectorOptions(
                    dimension=2,
                ),
            ),
        )
        col.drop()
        assert "sync_collection_to_drop" not in sync_database.list_collection_names()

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.describe("test of database metainformation, sync")
    def test_get_database_info_sync(
        self,
        sync_database: Database,
        data_api_credentials_kwargs: DataAPICredentials,
    ) -> None:
        assert isinstance(sync_database.id, str)
        assert isinstance(sync_database.name(), str)
        assert sync_database.keyspace == data_api_credentials_kwargs["keyspace"]
        assert isinstance(sync_database.info(), AstraDBDatabaseInfo)
        assert isinstance(sync_database.info().raw, dict)

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.describe("test of collection metainformation, sync")
    def test_get_collection_info_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        info = sync_collection.info()
        assert info.keyspace == sync_collection.keyspace
        assert info.keyspace == sync_collection.database.keyspace

    @pytest.mark.describe("test of Database list_collections, sync")
    def test_database_list_collections_sync(
        self,
        sync_database: Database,
        sync_collection: DefaultCollection,
    ) -> None:
        assert TEST_COLLECTION_NAME in sync_database.list_collection_names()

    @pytest.mark.describe("test of Collection options, sync")
    def test_collection_options_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        options = sync_collection.options()
        assert options.vector is not None
        assert options.vector.dimension == 2

    @pytest.mark.skipif(
        SECONDARY_KEYSPACE is None, reason="No secondary keyspace provided"
    )
    @pytest.mark.describe("test of Database list_collections on cross-keyspaces, sync")
    def test_database_list_collections_cross_keyspace_sync(
        self,
        sync_database: Database,
        sync_collection: DefaultCollection,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        assert TEST_COLLECTION_NAME not in sync_database.list_collection_names(
            keyspace=data_api_credentials_info["secondary_keyspace"]
        )

    @pytest.mark.skipif(
        SECONDARY_KEYSPACE is None, reason="No secondary keyspace provided"
    )
    @pytest.mark.describe("test of Database use_keyspace, sync")
    def test_database_use_keyspace_sync(
        self,
        sync_database: Database,
        sync_collection: DefaultCollection,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        # make a copy to avoid mutating the fixture
        t_database = sync_database._copy()
        assert t_database == sync_database
        assert t_database.keyspace == data_api_credentials_kwargs["keyspace"]
        assert TEST_COLLECTION_NAME in t_database.list_collection_names()

        t_database.use_keyspace(data_api_credentials_info["secondary_keyspace"])  # type: ignore[arg-type]
        assert t_database != sync_database
        assert t_database.keyspace == data_api_credentials_info["secondary_keyspace"]
        assert TEST_COLLECTION_NAME not in t_database.list_collection_names()

    @pytest.mark.skipif(
        SECONDARY_KEYSPACE is None, reason="No secondary keyspace provided"
    )
    @pytest.mark.describe("test of cross-keyspace collection lifecycle, sync")
    def test_collection_keyspace_sync(
        self,
        sync_database: Database,
        client: DataAPIClient,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        TEST_LOCAL_COLLECTION_NAME1 = "test_crossks_coll1"
        TEST_LOCAL_COLLECTION_NAME2 = "test_crossks_coll2"
        database_on_secondary = client.get_database(
            data_api_credentials_kwargs["api_endpoint"],
            token=data_api_credentials_kwargs["token"],
            keyspace=data_api_credentials_info["secondary_keyspace"],
        )
        sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME1,
            keyspace=data_api_credentials_info["secondary_keyspace"],
        )
        col2_on_secondary = sync_database.create_collection(
            TEST_LOCAL_COLLECTION_NAME2,
            keyspace=data_api_credentials_info["secondary_keyspace"],
        )
        assert (
            TEST_LOCAL_COLLECTION_NAME1 in database_on_secondary.list_collection_names()
        )
        database_on_secondary.drop_collection(TEST_LOCAL_COLLECTION_NAME1)
        sync_database.drop_collection(
            col2_on_secondary.name,
            keyspace=data_api_credentials_info["secondary_keyspace"],
        )
        assert (
            TEST_LOCAL_COLLECTION_NAME1
            not in database_on_secondary.list_collection_names()
        )
        assert (
            TEST_LOCAL_COLLECTION_NAME2
            not in database_on_secondary.list_collection_names()
        )

    @pytest.mark.describe("test of database command targeting collection, sync")
    def test_database_command_on_collection_sync(
        self,
        sync_database: Database,
        sync_collection: DefaultCollection,
    ) -> None:
        cmd1 = sync_database.command(
            {"countDocuments": {}}, collection_or_table_name=sync_collection.name
        )
        assert isinstance(cmd1, dict)
        assert isinstance(cmd1["status"]["count"], int)
        cmd2 = sync_database._copy(keyspace="...").command(
            {"countDocuments": {}},
            keyspace=sync_collection.keyspace,
            collection_or_table_name=sync_collection.name,
        )
        assert cmd2 == cmd1
        assert "count" in (cmd2.get("status") or {})

    @pytest.mark.describe("test of database command, sync")
    def test_database_command_sync(
        self,
        sync_database: Database,
    ) -> None:
        cmd1 = sync_database.command({"findCollections": {}})
        assert isinstance(cmd1, dict)
        assert isinstance(cmd1["status"]["collections"], list)
        cmd2 = sync_database._copy(keyspace="...").command(
            {"findCollections": {}}, keyspace=sync_database.keyspace
        )
        assert cmd2 == cmd1
        assert "collections" in (cmd2.get("status") or {})

        cmd_feps = sync_database.command(
            {"findEmbeddingProviders": {}},
            keyspace=None,
        )
        assert "embeddingProviders" in (cmd_feps.get("status") or {})

    @pytest.mark.describe("test of database command, sync")
    def test_collection_command_sync(
        self,
        sync_collection: DefaultCollection,
    ) -> None:
        cmd1 = sync_collection.command({"countDocuments": {}})
        assert isinstance(cmd1, dict)
        assert isinstance(cmd1["status"]["count"], int)

    @pytest.mark.describe("test of tokenless client creation, sync")
    def test_tokenless_client_sync(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        api_endpoint = data_api_credentials_kwargs["api_endpoint"]
        token = data_api_credentials_kwargs["token"]
        client = DataAPIClient(environment=data_api_credentials_info["environment"])
        database = client.get_database(
            api_endpoint,
            token=token,
            keyspace=data_api_credentials_kwargs["keyspace"],
        )
        assert isinstance(database.list_collection_names(), list)

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.describe(
        "test database-from-admin default keyspace per environment, sync"
    )
    def test_database_from_admin_default_keyspace_per_environment_sync(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        client = DataAPIClient(environment=data_api_credentials_info["environment"])
        admin = client.get_admin(token=data_api_credentials_kwargs["token"])
        db_m = admin.get_database(
            data_api_credentials_kwargs["api_endpoint"],
            keyspace="M",
        )
        assert db_m.keyspace == "M"
        db_n = admin.get_database(data_api_credentials_kwargs["api_endpoint"])
        assert isinstance(db_n.keyspace, str)  # i.e. resolution took place

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.describe(
        "test database-from-astradbadmin default keyspace per environment, sync"
    )
    def test_database_from_astradbadmin_default_keyspace_per_environment_sync(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        client = DataAPIClient(environment=data_api_credentials_info["environment"])
        admin = client.get_admin(token=data_api_credentials_kwargs["token"])
        db_admin = admin.get_database_admin(data_api_credentials_kwargs["api_endpoint"])
        db_m = db_admin.get_database(keyspace="M")
        assert db_m.keyspace == "M"
        db_n = db_admin.get_database()
        assert isinstance(db_n.keyspace, str)  # i.e. resolution took place

    @pytest.mark.describe("test of collection find-and-rerank lifecycle, sync")
    def test_collection_farr_lifecycle_sync(
        self,
        sync_database: Database,
    ) -> None:
        try:
            TEST_FARR_COLLECTION_NAME = "test_farr_coll"
            col1 = sync_database.create_collection(
                TEST_FARR_COLLECTION_NAME,
                definition=CollectionDefinition(
                    vector=CollectionVectorOptions(
                        dimension=10,
                        metric=VectorMetric.EUCLIDEAN,
                    ),
                    rerank=CollectionRerankOptions(
                        service=RerankServiceOptions(
                            provider="nvidia",
                            model_name="nvidia/llama-3.2-nv-rerankqa-1b-v2",
                        ),
                    ),
                    lexical=CollectionLexicalOptions(
                        analyzer="STANDARD",
                    ),
                ),
            )
            # test other creation methods (no-op since just created)
            col1_dict = sync_database.create_collection(
                TEST_FARR_COLLECTION_NAME,
                definition={
                    "vector": {
                        "dimension": 10,
                        "metric": VectorMetric.EUCLIDEAN,
                    },
                    "rerank": {
                        "enabled": True,
                        "service": {
                            "provider": "nvidia",
                            "modelName": "nvidia/llama-3.2-nv-rerankqa-1b-v2",
                        },
                    },
                    "lexical": {
                        "enabled": True,
                        "analyzer": "STANDARD",
                    },
                },
            )
            assert col1_dict == col1
            fluent_definition1 = (
                CollectionDefinition.builder()
                .set_vector_dimension(10)
                .set_vector_metric(VectorMetric.EUCLIDEAN)
                .set_rerank("nvidia", "nvidia/llama-3.2-nv-rerankqa-1b-v2")
                .set_lexical("STANDARD")
                .build()
            )
            col1_fluent = sync_database.create_collection(
                TEST_FARR_COLLECTION_NAME,
                definition=fluent_definition1,
            )
            assert col1_fluent == col1

            lc_response = list(sync_database.list_collections())

            expected_coll_descriptor = CollectionDescriptor._from_dict(
                {
                    "name": TEST_FARR_COLLECTION_NAME,
                    "options": {
                        "vector": {
                            "dimension": 10,
                            "metric": "euclidean",
                            "sourceModel": "other",
                        },
                        "rerank": {
                            "enabled": True,
                            "service": {
                                "provider": "nvidia",
                                "modelName": "nvidia/llama-3.2-nv-rerankqa-1b-v2",
                            },
                        },
                        "lexical": {
                            "enabled": True,
                            "analyzer": "STANDARD",
                        },
                    },
                },
            )
            assert expected_coll_descriptor in lc_response

            col2 = sync_database.get_collection(TEST_FARR_COLLECTION_NAME)
            assert col1 == col2
        finally:
            sync_database.drop_collection(TEST_FARR_COLLECTION_NAME)

    @pytest.mark.describe("test of collection lexical detailed config, sync")
    def test_collection_lexical_detailedconfig_sync(
        self,
        sync_database: Database,
    ) -> None:
        TEST_LEXICAL_DETAILEDCONFIG_COLLECTION_NAME = "test_lexical_detcfg_coll"
        try:
            sync_database.create_collection(
                TEST_LEXICAL_DETAILEDCONFIG_COLLECTION_NAME,
                definition=(
                    CollectionDefinition.builder()
                    .set_lexical(
                        {
                            "tokenizer": {"name": "standard", "args": {}},
                            "filters": [
                                {"name": "lowercase"},
                                {"name": "stop"},
                                {"name": "porterstem"},
                                {"name": "asciifolding"},
                            ],
                            "charFilters": [],
                        }
                    )
                    .build()
                ),
            )

            coll_ldc_definition = CollectionDefinition(
                lexical=CollectionLexicalOptions(
                    analyzer={
                        "tokenizer": {"name": "standard", "args": {}},
                        "filters": [
                            {"name": "lowercase"},
                            {"name": "stop"},
                            {"name": "porterstem"},
                            {"name": "asciifolding"},
                        ],
                        "charFilters": [],
                    },
                    enabled=True,
                ),
            )
            sync_database.create_collection(
                TEST_LEXICAL_DETAILEDCONFIG_COLLECTION_NAME,
                definition=coll_ldc_definition,
            )

        finally:
            sync_database.drop_collection(TEST_LEXICAL_DETAILEDCONFIG_COLLECTION_NAME)

    @pytest.mark.describe("test of collection lexical omitted settings, sync")
    def test_collection_lexical_omittedsettings_sync(
        self,
        sync_database: Database,
    ) -> None:
        TEST_LEXICAL_OMITTEDSETTINGS_COLLECTION_NAME = "test_lexical_nocfg_coll"
        try:
            sync_database.create_collection(
                TEST_LEXICAL_OMITTEDSETTINGS_COLLECTION_NAME,
                definition=(
                    CollectionDefinition.builder().set_lexical(enabled=True).build()
                ),
            )

            coll_ldc_definition = CollectionDefinition(
                lexical=CollectionLexicalOptions(enabled=True),
            )
            sync_database.create_collection(
                TEST_LEXICAL_OMITTEDSETTINGS_COLLECTION_NAME,
                definition=coll_ldc_definition,
            )

        finally:
            sync_database.drop_collection(TEST_LEXICAL_OMITTEDSETTINGS_COLLECTION_NAME)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.api_options import APIOptions, SerdesOptions

from ..conftest import DefaultAsyncTable, _repaint_NaNs
from .table_row_assets import (
    FULL_AR_ROW_CUSTOMTYPED,
    FULL_AR_ROW_NONCUSTOMTYPED,
)


class TestTableColumnTypesSync:
    @pytest.mark.describe("test of table write-and-read, with custom types, async")
    async def test_table_write_then_read_customtypes_async(
        self,
        async_empty_table_all_returns: DefaultAsyncTable,
    ) -> None:
        cdtypes_atable = async_empty_table_all_returns.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            ),
        )
        await cdtypes_atable.insert_one(FULL_AR_ROW_CUSTOMTYPED)
        retrieved = await cdtypes_atable.find_one({})
        assert _repaint_NaNs(retrieved) == _repaint_NaNs(FULL_AR_ROW_CUSTOMTYPED)

    @pytest.mark.describe("test of table write-and-read, with noncustom types, async")
    async def test_table_write_then_read_noncustomtypes_async(
        self,
        async_empty_table_all_returns: DefaultAsyncTable,
    ) -> None:
        rdtypes_atable = async_empty_table_all_returns.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            ),
        )
        await rdtypes_atable.insert_one(FULL_AR_ROW_NONCUSTOMTYPED)
        retrieved = await rdtypes_atable.find_one({})
        assert _repaint_NaNs(retrieved) == _repaint_NaNs(FULL_AR_ROW_NONCUSTOMTYPED)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.data.info.table_descriptor.table_altering import AlterTableOperation
from astrapy.data.info.table_descriptor.type_altering import AlterTypeOperation
from astrapy.data.utils.table_types import ColumnType
from astrapy.info import (
    AlterTableAddColumns,
    AlterTableAddVectorize,
    AlterTableDropColumns,
    AlterTableDropVectorize,
    AlterTypeAddFields,
    AlterTypeRenameFields,
    CreateTableDefinition,
    CreateTypeDefinition,
    ListTableDefinition,
    ListTableDescriptor,
    TableAPIIndexSupportDescriptor,
    TableIndexDefinition,
    TableIndexOptions,
    TablePrimaryKeyDescriptor,
    TableScalarColumnTypeDescriptor,
    TableTextIndexDefinition,
    TableTextIndexOptions,
    TableUDTColumnDescriptor,
    TableUnsupportedIndexDefinition,
    TableVectorIndexDefinition,
    TableVectorIndexOptions,
    VectorServiceOptions,
)

TABLE_DICTS = [
    {
        "name": "table_simple",
        "definition": {
            "columns": {
                "country": {"type": "text"},
                "name": {"type": "text"},
                "human": {"type": "boolean"},
                "email": {"type": "text"},
                "age": {"type": "int"},
            },
            "primaryKey": {"partitionBy": ["email"], "partitionSort": {}},
        },
    },
    {
        "name": "with_a_vector",
        "definition": {
            "columns": {
                "city": {"type": "text"},
                "name": {"type": "text"},
                "id": {"type": "text"},
                "age": {"type": "int"},
                "somevector": {"type": "vector", "dimension": 123},
            },
            "primaryKey": {"partitionBy": ["id"], "partitionSort": {}},
        },
    },
    {
        "name": "with_vectorize",
        "definition": {
            "columns": {
                "embeddings": {
                    "type": "vector",
                    "dimension": 123,
                    "service": {
                        "provider": "openai",
                        "modelName": "text-embedding-3-small",
                    },
                },
                "city": {"type": "text"},
                "name": {"type": "text"},
                "id": {"type": "text"},
                "age": {"type": "int"},
            },
            "primaryKey": {"partitionBy": ["id"], "partitionSort": {}},
        },
    },
    {
        "name": "with_uncreatable_columns",
        "definition": {
            "columns": {
                "p_text": {"type": "text"},
                "p_counter": {
                    "type": "counter",
                    "apiSupport": {
                        "createTable": False,
                        "insert": False,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "counter",
                    },
                },
                "p_timeuuid": {
                    "type": "timeuuid",
                    "apiSupport": {
                        "createTable": False,
                        "insert": True,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "timeuuid",
                    },
                },
            },
            "primaryKey": {"partitionBy": ["p_text"], "partitionSort": {}},
        },
    },
    {
        "name": "with_unsupported",
        "definition": {
            "columns": {
                "p_text": {"type": "text"},
                "my_nested_frozen": {
                    "type": "UNSUPPORTED",
                    "apiSupport": {
                        "createTable": False,
                        "insert": False,
                        "read": False,
                        "filter": False,
                        "cqlDefinition": "map<text, frozen<list<int>>>",
                    },
                },
            },
            "primaryKey": {"partitionBy": ["p_text"], "partitionSort": {}},
        },
    },
    {
        "name": "without_apisupports",
        "definition": {
            "columns": {
                "p_scalar": {
                    "type": "text",
                },
                "p_list": {
                    "type": "list",
                    "valueType": {"type": "text"},
                },
                "p_map": {
                    "type": "map",
                    "keyType": {"type": "text"},
                    "valueType": {"type": "text"},
                },
                "p_vector": {
                    "type": "vector",
                    "dimension": 999,
                },
                "p_unsupported": {
                    "type": "UNSUPPORTED",
                    "apiSupport": {
                        "createTable": False,
                        "insert": False,
                        "read": False,
                        "filter": False,
                        "cqlDefinition": "real unsupported cannot omit apiSupport!",
                    },
                },
            },
            "primaryKey": {"partitionBy": ["id"], "partitionSort": {}},
        },
    },
    {
        "name": "with_all_apisupports",
        "definition": {
            "columns": {
                "p_scalar": {
                    "type": "text",
                    "apiSupport": {
                        "createTable": True,
                        "insert": True,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "for scalar, not seen in API yet",
                    },
                },
                "p_list": {
                    "type": "list",
                    "valueType": {"type": "text"},
                    "apiSupport": {
                        "createTable": True,
                        "insert": True,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "the API returns apisupport for this type",
                    },
                },
                "p_map": {
                    "type": "map",
                    "keyType": {"type": "text"},
                    "valueType": {"type": "text"},
                    "apiSupport": {
                        "createTable": True,
                        "insert": True,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "the API returns apisupport for this type",
                    },
                },
                "p_vector": {
                    "type": "vector",
                    "dimension": 999,
                    "apiSupport": {
                        "createTable": True,
                        "insert": True,
                        "read": True,
                        "filter": True,
                        "cqlDefinition": "the API returns apisupport for this type",
                    },
                },
                "p_unsupported": {
                    "type": "UNSUPPORTED",
                    "apiSupport": {
                        "createTable": False,
                        "insert": False,
                        "read": False,
                        "filter": False,
                        "cqlDefinition": "real unsupported cannot omit apiSupport!",
                    },
                },
            },
            "primaryKey": {"partitionBy": ["id"], "partitionSort": {}},
        },
    },
]

DICT_DEFINITION = {
    "columns": {
        "country": {"type": "text"},
        "name": {"type": "text"},
        "human": {"type": "boolean"},
        "email": {"type": "text"},
        "age": {"type": "int"},
    },
    "primaryKey": {"partitionBy": ["email"], "partitionSort": {}},
}
HYBRID_DEFINITION = {
    "columns": {
        "country": TableScalarColumnTypeDescriptor(column_type="text"),
        "name": {"type": "text"},
        "human": TableScalarColumnTypeDescriptor(column_type="boolean"),
        "email": {"type": "text"},
        "age": TableScalarColumnTypeDescriptor(column_type="int"),
    },
    "primaryKey": TablePrimaryKeyDescriptor(
        partition_by=["email"],
        partition_sort={},
    ),
}

SHORT_FORM_COLUMN_TYPE = "int"
LONG_FORM_COLUMN_TYPE = {"type": "int"}

SHORT_FORM_PRIMARY_KEY = "column"
LONG_FORM_PRIMARY_KEY = {"partitionBy": ["column"], "partitionSort": {}}

INDEX_DEFINITION_FULL = TableIndexDefinition(
    column="the_column",
    options=TableIndexOptions(
        ascii=True,
        normalize=False,
        case_sensitive=True,
    ),
)
INDEX_DEFINITION_DICT_FULL = {
    "column": "the_column",
    "options": {
        "ascii": True,
        "normalize": False,
        "caseSensitive": True,
    },
}
INDEX_DEFINITION_DICT_PARTIAL = {
    "column": "the_column",
    "options": {
        "normalize": True,
    },
}
INDEX_DEFINITION_DICT_MINIMAL = {"column": "the_column"}
INDEX_DEFINITION_DICT_COERCEABLE_MINIMAL = {"column": "the_column"}

VECTOR_INDEX_DEFINITION_FULL = TableVectorIndexDefinition(
    column="the_v_column",
    options=TableVectorIndexOptions(
        metric="the_metric",
        source_model="the_source_model",
    ),
)
VECTOR_INDEX_DEFINITION_DICT_FULL = {
    "column": "the_v_column",
    "options": {
        "metric": "the_metric",
        "sourceModel": "the_source_model",
    },
}
VECTOR_INDEX_DEFINITION_DICT_PARTIAL = {
    "column": "the_v_column",
    "options": {
        "sourceModel": "the_source_model",
    },
}
VECTOR_INDEX_DEFINITION_DICT_MINIMAL = {"column": "the_v_column"}
VECTOR_INDEX_DEFINITION_DICT_COERCEABLE_MINIMAL = {"column": "the_v_column"}

TEXT_INDEX_DEFINITION_SHORT_DICT = {
    "column": "the_analyz_column1",
    "options": {
        "analyzer": "STANDARD",
    },
}
TEXT_INDEX_DEFINITION_SHORT = TableTextIndexDefinition(
    column="the_analyz_column1",
    options=TableTextIndexOptions(analyzer="STANDARD"),
)
TEXT_INDEX_DEFINITION_LONG_DICT = {
    "column": "the_analyz_column1",
    "options": {"analyzer": {"tokenizer": {"name": "whitespace"}}},
}
TEXT_INDEX_DEFINITION_LONG = TableTextIndexDefinition(
    column="the_analyz_column1",
    options=TableTextIndexOptions(
        analyzer={
            "tokenizer": {
                "name": "whitespace",
            },
        }
    ),
)

UNSUPPORTED_INDEX_DEFINITION = TableUnsupportedIndexDefinition(
    column="UNKNOWN",
    api_support=TableAPIIndexSupportDescriptor(
        create_index=False,
        filter=True,
        cql_definition="CREATE INDEX SO-AND-SO!",
    ),
)
UNSUPPORTED_INDEX_DEFINITION_DICT = {
    "column": "UNKNOWN",
    "apiSupport": {
        "createIndex": False,
        "filter": True,
        "cqlDefinition": "CREATE INDEX SO-AND-SO!",
    },
}


class TestListTableDescriptors:
    @pytest.mark.describe("test of parsing list-table descriptors, fully dict")
    def test_listtabledescriptor_parsing_fulldict(self) -> None:
        table_descs = [
            ListTableDescriptor.coerce(table_dict) for table_dict in TABLE_DICTS
        ]
        assert all(
            table_desc.as_dict() == table_dict
            for table_desc, table_dict in zip(table_descs, TABLE_DICTS)
        )

    @pytest.mark.describe("test of parsing list-table descriptors with hybrid inputs")
    def test_listtabledescriptor_hybrid_parsing(self) -> None:
        from_dict_def = ListTableDefinition.coerce(DICT_DEFINITION)
        from_hyb_def = ListTableDefinition.coerce(HYBRID_DEFINITION)
        assert from_dict_def == from_hyb_def

    @pytest.mark.describe("test of parsing short forms for column types")
    def test_columntype_short_form(self) -> None:
        long_col = TableScalarColumnTypeDescriptor.coerce(LONG_FORM_COLUMN_TYPE)
        short_col = TableScalarColumnTypeDescriptor.coerce(SHORT_FORM_COLUMN_TYPE)
        assert long_col == short_col

    @pytest.mark.describe("test of parsing short forms for primary-key descriptor")
    def test_primarykey_short_form(self) -> None:
        long_pk = TablePrimaryKeyDescriptor.coerce(LONG_FORM_PRIMARY_KEY)
        short_pk = TablePrimaryKeyDescriptor.coerce(SHORT_FORM_PRIMARY_KEY)
        assert long_pk == short_pk

    @pytest.mark.describe("test of fluent interface for table definition")
    def test_tabledefinition_fluent(self) -> None:
        pb0 = (
            CreateTableDefinition.builder()
            .add_partition_by(["p_text", "p_int"])
            .build()
        )
        pb1 = (
            CreateTableDefinition.builder()
            .add_partition_by("p_text")
            .add_partition_by("p_int")
            .build()
        )
        pb2 = (
            CreateTableDefinition.builder()
            .add_partition_by(["p_text"])
            .add_partition_by(["p_int"])
            .build()
        )
        pb3 = (
            CreateTableDefinition.builder()
            .add_partition_by("p_text")
            .add_partition_by(["p_int"])
            .build()
        )
        pbz = (
            CreateTableDefinition.builder()
            .add_partition_by(["p_int"])
            .add_partition_by("p_text")
            .build()
        )
        assert pb0 == pb1
        assert pb0 == pb2
        assert pb0 == pb3
        assert pb0 != pbz

        def0 = (
            CreateTableDefinition.builder()
            .add_column("p_text", "text")
            .add_column("p_int", "int")
            .add_column("p_boolean", "boolean")
            .add_scalar_column("p_float", "float")
            .add_set_column("p_set", ColumnType.INT)
            .add_list_column("p_list", ColumnType.FLOAT)
            .add_map_column(
                "p_map",
                {"type": "text"},
                TableScalarColumnTypeDescriptor(column_type=ColumnType.INT),
            )
            .add_vector_column("p_vector", dimension=191)
            .add_vector_column(
                "p_vectorize",
                dimension=1024,
                service=VectorServiceOptions(
                    provider="mistral",
                    model_name="mistral-embed",
                ),
            )
            .add_userdefinedtype_column("p_udt", "the_udt")
            .add_set_column(
                "p_set_udt", {"type": "userDefined", "udtName": "the_udt_s"}
            )
            .add_list_column(
                "p_list_udt",
                TableUDTColumnDescriptor(udt_name="the_udt_l"),
            )
            .add_map_column(
                "p_map_udt",
                {"type": "text"},
                {"type": "userDefined", "udtName": "the_udt_m"},
            )
            .add_partition_by(["p_text", "p_int"])
            .add_partition_sort({"p_boolean": -1, "p_float": 1})
            .build()
        )
        def1 = CreateTableDefinition.coerce(
            {
                "columns": {
                    "p_text": {"type": "text"},
                    "p_int": {"type": "int"},
                    "p_boolean": {"type": "boolean"},
                    "p_float": {"type": "float"},
                    "p_set": {"type": "set", "valueType": "int"},
                    "p_list": {"type": "list", "valueType": "float"},
                    "p_map": {"type": "map", "keyType": "text", "valueType": "int"},
                    "p_vector": {"type": "vector", "dimension": 191},
                    "p_vectorize": {
                        "type": "vector",
                        "dimension": 1024,
                        "service": {
                            "provider": "mistral",
                            "modelName": "mistral-embed",
                        },
                    },
                    "p_udt": {
                        "type": "userDefined",
                        "udtName": "the_udt",
                    },
                    "p_set_udt": {
                        "type": "set",
                        "valueType": {
                            "type": "userDefined",
                            "udtName": "the_udt_s",
                        },
                    },
                    "p_list_udt": {
                        "type": "list",
                        "valueType": {
                            "type": "userDefined",
                            "udtName": "the_udt_l",
                        },
                    },
                    "p_map_udt": {
                        "type": "map",
                        "keyType": "text",
                        "valueType": {
                            "type": "userDefined",
                            "udtName": "the_udt_m",
                        },
                    },
                },
                "primaryKey": {
                    "partitionBy": [
                        "p_text",
                        "p_int",
                    ],
                    "partitionSort": {"p_boolean": -1, "p_float": 1},
                },
            }
        )
        assert def0 == def1

        adef0 = (
            CreateTableDefinition.builder()
            .add_list_column("p_list", "tinyint")
            .add_vector_column(
                "p_vectorize",
                dimension=333,
                service={
                    "provider": "mistral",
                    "modelName": "mistral-embed",
                },
            )
            .build()
        )
        adef1 = CreateTableDefinition.coerce(
            {
                "columns": {
                    "p_list": {"type": "list", "valueType": "tinyint"},
                    "p_vectorize": {
                        "type": "vector",
                        "dimension": 333,
                        "service": {
                            "provider": "mistral",
                            "modelName": "mistral-embed",
                        },
                    },
                },
                "primaryKey": {"partitionBy": [], "partitionSort": {}},
            }
        )
        assert adef0 == adef1

    @pytest.mark.describe("test of parsing of index definitions")
    def test_indexdefinition_parsing(self) -> None:
        ti_full = TableIndexDefinition.coerce(
            INDEX_DEFINITION_DICT_FULL,
        )
        assert INDEX_DEFINITION_DICT_FULL == ti_full.as_dict()
        assert ti_full == INDEX_DEFINITION_FULL

        ti_partial = TableIndexDefinition.coerce(
            INDEX_DEFINITION_DICT_PARTIAL,
        )
        assert INDEX_DEFINITION_DICT_PARTIAL == ti_partial.as_dict()

        ti_minimal = TableIndexDefinition.coerce(
            INDEX_DEFINITION_DICT_MINIMAL,
        )
        assert INDEX_DEFINITION_DICT_MINIMAL == ti_minimal.as_dict()

        ti_coerceable_minimal = TableIndexDefinition.coerce(
            INDEX_DEFINITION_DICT_COERCEABLE_MINIMAL,
        )
        assert INDEX_DEFINITION_DICT_MINIMAL == ti_coerceable_minimal.as_dict()

    @pytest.mark.describe("test of parsing of map index definitions")
    def test_map_indexdefinition_parsing(self) -> None:
        assert TableIndexDefinition.coerce(
            {"column": "the_text"}
        ) == TableIndexDefinition(column="the_text", options=TableIndexOptions())

        assert TableIndexDefinition.coerce(
            {"column": {"the_map": "$keys"}}
        ) == TableIndexDefinition(
            column={"the_map": "$keys"}, options=TableIndexOptions()
        )

        assert TableIndexDefinition.coerce(
            {"column": {"the_map": "$values"}}
        ) == TableIndexDefinition(
            column={"the_map": "$values"}, options=TableIndexOptions()
        )

    @pytest.mark.describe("test of parsing of vector index definitions")
    def test_vectorindexdefinition_parsing(self) -> None:
        tvi_full = TableVectorIndexDefinition.coerce(
            VECTOR_INDEX_DEFINITION_DICT_FULL,
        )
        assert VECTOR_INDEX_DEFINITION_DICT_FULL == tvi_full.as_dict()
        assert tvi_full == VECTOR_INDEX_DEFINITION_FULL

        tvi_partial = TableVectorIndexDefinition.coerce(
            VECTOR_INDEX_DEFINITION_DICT_PARTIAL,
        )
        assert VECTOR_INDEX_DEFINITION_DICT_PARTIAL == tvi_partial.as_dict()
        tvi_minimal = TableVectorIndexDefinition.coerce(
            VECTOR_INDEX_DEFINITION_DICT_MINIMAL,
        )
        assert VECTOR_INDEX_DEFINITION_DICT_MINIMAL == tvi_minimal.as_dict()

        tvi_coerceable_minimal = TableVectorIndexDefinition.coerce(
            VECTOR_INDEX_DEFINITION_DICT_COERCEABLE_MINIMAL,
        )
        assert VECTOR_INDEX_DEFINITION_DICT_MINIMAL == tvi_coerceable_minimal.as_dict()

    @pytest.mark.describe("test of parsing of text index definitions")
    def test_textindexdefinition_parsing(self) -> None:
        tti_short = TableTextIndexDefinition.coerce(
            TEXT_INDEX_DEFINITION_SHORT_DICT,
        )
        assert TEXT_INDEX_DEFINITION_SHORT_DICT == tti_short.as_dict()
        assert tti_short == TEXT_INDEX_DEFINITION_SHORT

        tti_long = TableTextIndexDefinition.coerce(
            TEXT_INDEX_DEFINITION_LONG_DICT,
        )
        assert TEXT_INDEX_DEFINITION_LONG_DICT == tti_long.as_dict()
        assert tti_long == TEXT_INDEX_DEFINITION_LONG

    @pytest.mark.describe("test of parsing of unsupported index definitions")
    def test_unsupportedindexdefinition_parsing(self) -> None:
        tui_full = TableUnsupportedIndexDefinition.coerce(
            UNSUPPORTED_INDEX_DEFINITION_DICT,
        )
        assert UNSUPPORTED_INDEX_DEFINITION_DICT == tui_full.as_dict()
        assert tui_full == UNSUPPORTED_INDEX_DEFINITION

    @pytest.mark.describe("test of parsing AlterTableOperation classes")
    def test_altertableoperation_parsing(self) -> None:
        addc_o = AlterTableOperation.from_full_dict(
            {
                "add": {
                    "columns": {
                        "p_int": {"type": "int"},
                        "p_text": "text",
                    }
                },
            },
        )
        addc = AlterTableAddColumns(
            columns={
                "p_int": TableScalarColumnTypeDescriptor(column_type=ColumnType.INT),
                "p_text": TableScalarColumnTypeDescriptor(column_type=ColumnType.TEXT),
            }
        )
        assert addc_o == addc

        addv_o = AlterTableOperation.from_full_dict(
            {
                "addVectorize": {
                    "columns": {"col_v": {"provider": "p", "modelName": "mn"}}
                }
            }
        )
        addv = AlterTableAddVectorize(
            columns={
                "col_v": VectorServiceOptions(
                    provider="p",
                    model_name="mn",
                )
            }
        )
        assert addv_o == addv

        dropc_o = AlterTableOperation.from_full_dict(
            {"drop": {"columns": ["col1", "col2"]}},
        )
        dropc = AlterTableDropColumns(columns=["col1", "col2"])
        assert dropc_o == dropc

        dropv_o = AlterTableOperation.from_full_dict(
            {"dropVectorize": {"columns": ["col1", "col2"]}},
        )
        dropv = AlterTableDropVectorize(columns=["col1", "col2"])
        assert dropv_o == dropv

    @pytest.mark.describe(
        "test of coerce normalizing column types for AlterTableAddColumns"
    )
    def test_altertableaddcolumns_normalizetypes(self) -> None:
        short_form_dict = {
            "columns": {
                "col1": "int",
                "col2": "text",
            }
        }
        normalized_dict = {
            "columns": {
                "col1": {"type": "int"},
                "col2": {"type": "text"},
            }
        }
        alter_table_add_cols = AlterTableAddColumns._from_dict(short_form_dict)
        assert alter_table_add_cols.as_dict() == normalized_dict

    @pytest.mark.parametrize(
        ("operation_class", "test_dict"),
        [
            (
                AlterTableAddColumns,
                {
                    "columns": {
                        "col1": {"type": "int"},
                        "col2": {"type": "text"},
                    }
                },
            ),
            (
                AlterTableAddVectorize,
                {
                    "columns": {
                        "vec_col": {
                            "provider": "openai",
                            "modelName": "text-embedding-3-small",
                        }
                    }
                },
            ),
            (
                AlterTableDropColumns,
                {
                    "columns": ["col1", "col2"],
                },
            ),
            (
                AlterTableDropVectorize,
                {
                    "columns": ["vec_col1", "vec_col2"],
                },
            ),
        ],
        ids=["addcolumns", "addvectorize", "dropcolumns", "dropvectorize"],
    )
    @pytest.mark.describe(
        "test of coerce and as_dict for alter table operation classes"
    )
    def test_altertableoperation_coerce_asdict(
        self,
        operation_class: type[AlterTableOperation],
        test_dict: dict[str, Any],
    ) -> None:
        # Test coerce with dict input
        operation_from_dict = operation_class.coerce(test_dict)

        # Test coerce with object input (should return same object)
        operation_from_obj = operation_class.coerce(operation_from_dict)
        assert operation_from_obj == operation_from_dict

        # Test as_dict returns the original dict
        assert operation_from_dict.as_dict() == test_dict

    @pytest.mark.describe("test of AlterTypeOperation classes")
    def test_altertypeoperation_parsing_coerce_asdict(self) -> None:
        addf = AlterTypeAddFields(
            fields={
                "p_text": TableScalarColumnTypeDescriptor(column_type=ColumnType.TEXT),
                "p_int": TableScalarColumnTypeDescriptor(column_type=ColumnType.INT),
            }
        )

        # constructor signatures
        addf_constr_simple = AlterTypeAddFields(
            fields={
                "p_text": "text",
                "p_int": "int",
            }
        )
        addf_constr_dicts = AlterTypeAddFields(
            fields={
                "p_text": {"type": "text"},
                "p_int": {"type": "int"},
            }
        )

        assert addf == addf_constr_simple
        assert addf == addf_constr_dicts

        # other ways to construct
        addf_o = AlterTypeOperation.from_full_dict(
            {"add": {"fields": {"p_text": "text", "p_int": {"type": "int"}}}},
        )
        assert addf_o == addf

        assert AlterTypeAddFields.coerce(addf.as_dict()) == addf

        renf_o = AlterTypeOperation.from_full_dict(
            {"rename": {"fields": {"p_text": "new_p_text", "p_int": "new_p_int"}}},
        )
        renf = AlterTypeRenameFields(
            fields={
                "p_text": "new_p_text",
                "p_int": "new_p_int",
            }
        )
        assert renf_o == renf

        assert AlterTypeRenameFields.coerce(renf.as_dict()) == renf

    @pytest.mark.parametrize(
        ["ops_to_stack", "expected_op", "op_class"],
        [
            (
                [
                    AlterTypeAddFields(fields={"f_text": "blob", "f_int": "int"}),
                    AlterTypeAddFields(fields={"f_float": "blob"}),
                    AlterTypeAddFields(fields={"f_text": "text", "f_float": "float"}),
                ],
                AlterTypeAddFields(
                    fields={
                        "f_text": "text",
                        "f_int": "int",
                        "f_float": "float",
                    }
                ),
                AlterTypeAddFields,
            ),
            (
                [
                    AlterTypeRenameFields(fields={"f_text": "x", "f_int": "F_INT"}),
                    AlterTypeRenameFields(fields={"f_float": "x"}),
                    AlterTypeRenameFields(
                        fields={"f_text": "F_TEXT", "f_float": "F_FLOAT"}
                    ),
                ],
                AlterTypeRenameFields(
                    fields={
                        "f_text": "F_TEXT",
                        "f_int": "F_INT",
                        "f_float": "F_FLOAT",
                    }
                ),
                AlterTypeRenameFields,
            ),
        ],
        ids=["add", "rename"],
    )
    @pytest.mark.describe("test of AlterTypeOperation class stacking")
    def test_altertypeoperation_stacking(
        self,
        ops_to_stack: list[AlterTypeOperation],
        expected_op: AlterTypeOperation,
        op_class: type[AlterTypeOperation],
    ) -> None:
        assert op_class.stack(ops_to_stack) == expected_op

    @pytest.mark.describe("test of CreateTypeDefinition")
    def test_createtypedefinition_parsing(self) -> None:
        type_definition_objects = CreateTypeDefinition(
            fields={
                "tagline": TableScalarColumnTypeDescriptor(ColumnType.TEXT),
                "score": TableScalarColumnTypeDescriptor(ColumnType.INT),
                "weight": TableScalarColumnTypeDescriptor(ColumnType.FLOAT),
            },
        )

        type_definition_fluent = (
            CreateTypeDefinition.builder()
            .add_field("tagline", ColumnType.TEXT)
            .add_field("score", "int")
            .add_field("weight", ColumnType.FLOAT)
            .build()
        )

        fields_dict_simple = {
            "fields": {
                "tagline": "text",
                "score": "int",
                "weight": "float",
            },
        }
        type_definition_simple = CreateTypeDefinition.coerce(fields_dict_simple)

        fields_dict_dicts = {
            "fields": {
                "tagline": {"type": "text"},
                "score": {"type": "int"},
                "weight": {"type": "float"},
            },
        }
        type_definition_dicts = CreateTypeDefinition.coerce(fields_dict_dicts)

        fields_dict_mixed = {
            "fields": {
                "tagline": TableScalarColumnTypeDescriptor(ColumnType.TEXT),
                "score": "int",
                "weight": {"type": "float"},
            },
        }
        type_definition_mixed = CreateTypeDefinition.coerce(fields_dict_mixed)

        assert type_definition_objects == type_definition_fluent
        assert type_definition_objects == type_definition_simple
        assert type_definition_objects == type_definition_dicts
        assert type_definition_objects == type_definition_mixed

        assert type_definition_objects.as_dict() == fields_dict_dicts

        assert (
            CreateTypeDefinition.coerce(type_definition_objects.as_dict())
            == type_definition_objects
        )
        assert (
            CreateTypeDefinition.coerce(type_definition_objects)
            == type_definition_objects
        )

    @pytest.mark.describe("test of AlterTypeOperation stack_by_name")
    def test_altertypeoperation_stackbybname(self) -> None:
        ops = [
            AlterTypeAddFields(fields={"a0": "text", "a1": {"type": "int"}}),
            AlterTypeRenameFields(fields={"b0": "rb0", "b1": "rb1"}),
            AlterTypeAddFields(fields={"a1": "int", "a2": "float", "a3": "timestamp"}),
            AlterTypeRenameFields(fields={"b1": "rz1", "b2": "rz2"}),
            AlterTypeAddFields(fields={"a2": "ascii"}),
        ]

        expected_op_map = {
            "add": AlterTypeAddFields(
                fields={
                    "a0": "text",
                    "a1": "int",
                    "a2": "ascii",
                    "a3": "timestamp",
                }
            ),
            "rename": AlterTypeRenameFields(
                fields={
                    "b0": "rb0",
                    "b1": "rz1",
                    "b2": "rz2",
                }
            ),
        }

        stacked_op_map = AlterTypeOperation.stack_by_name(ops)

        assert stacked_op_map == expected_op_map



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import logging
from typing import Any

import pytest

from astrapy.info import (
    TableAPIIndexSupportDescriptor,
    TableIndexDefinition,
    TableIndexDescriptor,
    TableIndexOptions,
    TableIndexType,
    TableUnsupportedIndexDefinition,
    TableVectorIndexDefinition,
    TableVectorIndexOptions,
)

LOCAL_R_RESPONSE = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "ascii": false,
                        "caseSensitive": true,
                        "normalize": false
                    }
                },
                "indexType": "regular"
            }
        ]
    }
}""")
LOCAL_R_RESPONSE_W = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "ascii": false,
                        "caseSensitive": true,
                        "normalize": false
                    }
                },
                "indexType": "WRONG_INDEX_TYPE"
            }
        ]
    }
}""")
PROD_R_RESPONSE = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "ascii": false,
                        "caseSensitive": true,
                        "normalize": false
                    }
                }
            }
        ]
    }
}""")

LOCAL_V_RESPONSE = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "metric": "cosine",
                        "sourceModel": "other"
                    }
                },
                "indexType": "vector"
            }
        ]
    }
}""")
LOCAL_V_RESPONSE_W = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "metric": "cosine",
                        "sourceModel": "other"
                    }
                },
                "indexType": "WRONG_INDEX_TYPE"
            }
        ]
    }
}""")
PROD_V_RESPONSE = json.loads("""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "b",
                    "options": {
                        "metric": "cosine",
                        "sourceModel": "other"
                    }
                }
            }
        ]
    }
}""")

LOCAL_U_RESPONSE = json.loads(r"""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "UNKNOWN",
                    "apiSupport": {
                        "createIndex": false,
                        "filter": false,
                        "cqlDefinition": "CREATE CUSTOM INDEX t1_b_index ON default_keyspace.t1 (\"bCamelCase\")\nUSING 'StorageAttachedIndex'\nWITH OPTIONS = {\n    'similarity_function' : 'COSINE',\n    'source_model' : 'OTHER'}"
                    }
                },
                "indexType": "UNKNOWN"
            }
        ]
    }
}""")
LOCAL_U_RESPONSE_W = json.loads(r"""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "UNKNOWN",
                    "apiSupport": {
                        "createIndex": false,
                        "filter": false,
                        "cqlDefinition": "CREATE CUSTOM INDEX t1_b_index ON default_keyspace.t1 (\"bCamelCase\")\nUSING 'StorageAttachedIndex'\nWITH OPTIONS = {\n    'similarity_function' : 'COSINE',\n    'source_model' : 'OTHER'}"
                    }
                },
                "indexType": "WRONG_INDEX_TYPE"
            }
        ]
    }
}""")
PROD_U_RESPONSE = json.loads(r"""{
    "status": {
        "indexes": [
            {
                "name": "t1_b_index",
                "definition": {
                    "column": "UNKNOWN",
                    "apiSupport": {
                        "createIndex": false,
                        "filter": false,
                        "cqlDefinition": "CREATE CUSTOM INDEX t1_b_index ON default_keyspace.t1 (\"bCamelCase\")\nUSING 'StorageAttachedIndex'\nWITH OPTIONS = {\n    'similarity_function' : 'COSINE',\n    'source_model' : 'OTHER'}"
                    }
                }
            }
        ]
    }
}""")


class TestTableIndexDescriptionParsing:
    @pytest.mark.describe("test of parsing index descriptions with/out indexType")
    @pytest.mark.parametrize(
        ("no_it_resp", "it_resp", "faulty_it_resp"),
        [
            (PROD_R_RESPONSE, LOCAL_R_RESPONSE, LOCAL_R_RESPONSE_W),
            (PROD_V_RESPONSE, LOCAL_V_RESPONSE, LOCAL_V_RESPONSE_W),
            (PROD_U_RESPONSE, LOCAL_U_RESPONSE, LOCAL_U_RESPONSE_W),
        ],
        ids=("regular", "vector", "unknown"),
    )
    def test_parsing_regular_index_description(
        self,
        caplog: pytest.LogCaptureFixture,
        no_it_resp: dict[str, Any],
        it_resp: dict[str, Any],
        faulty_it_resp: dict[str, Any],
    ) -> None:
        descriptors = [
            TableIndexDescriptor.coerce(response["status"]["indexes"][0])
            for response in [
                PROD_U_RESPONSE,
                LOCAL_U_RESPONSE,
            ]
        ]
        with caplog.at_level(logging.WARNING):
            descriptors.append(
                TableIndexDescriptor.coerce(LOCAL_U_RESPONSE_W["status"]["indexes"][0])
            )

        warnings = [rec for rec in caplog.records if rec.levelname == "WARNING"]
        assert len(warnings) == 1
        assert "WRONG_INDEX_TYPE" in warnings[0].message
        assert descriptors[0] == descriptors[1]
        assert descriptors[0] == descriptors[2]

    @pytest.mark.describe("test of creating regular TableIndexDescriptor instances")
    def test_create_reg_tableindexdescriptor_various_indextype(self) -> None:
        reg_ti_0 = TableIndexDescriptor(
            name="reg_ti",
            definition=TableIndexDefinition(
                column="reg_col",
                options=TableIndexOptions(
                    ascii=False,
                    normalize=True,
                    case_sensitive=False,
                ),
            ),
        )
        reg_ti_e = TableIndexDescriptor(
            name="reg_ti",
            definition=TableIndexDefinition(
                column="reg_col",
                options=TableIndexOptions(
                    ascii=False,
                    normalize=True,
                    case_sensitive=False,
                ),
            ),
            index_type=TableIndexType.REGULAR,
        )
        reg_ti_s = TableIndexDescriptor(
            name="reg_ti",
            definition=TableIndexDefinition(
                column="reg_col",
                options=TableIndexOptions(
                    ascii=False,
                    normalize=True,
                    case_sensitive=False,
                ),
            ),
            index_type="regular",
        )
        assert reg_ti_0 == reg_ti_e
        assert reg_ti_0 == reg_ti_s

    @pytest.mark.describe("test of creating vector TableIndexDescriptor instances")
    def test_create_vec_tableindexdescriptor_various_indextype(self) -> None:
        vec_ti_0 = TableIndexDescriptor(
            name="vec_ti",
            definition=TableVectorIndexDefinition(
                column="vec_col",
                options=TableVectorIndexOptions(metric="cosine"),
            ),
        )
        vec_ti_e = TableIndexDescriptor(
            name="vec_ti",
            definition=TableVectorIndexDefinition(
                column="vec_col",
                options=TableVectorIndexOptions(metric="cosine"),
            ),
            index_type=TableIndexType.VECTOR,
        )
        vec_ti_s = TableIndexDescriptor(
            name="vec_ti",
            definition=TableVectorIndexDefinition(
                column="vec_col",
                options=TableVectorIndexOptions(metric="cosine"),
            ),
            index_type="vector",
        )
        assert vec_ti_0 == vec_ti_e
        assert vec_ti_0 == vec_ti_s

    @pytest.mark.describe("test of creating unsupported TableIndexDescriptor instances")
    def test_create_uns_tableindexdescriptor_various_indextype(self) -> None:
        unk_ti_0 = TableIndexDescriptor(
            name="unk_ti",
            definition=TableUnsupportedIndexDefinition(
                column="UNKNOWN",
                api_support=TableAPIIndexSupportDescriptor(
                    cql_definition="c", create_index=False, filter=True
                ),
            ),
        )
        unk_ti_e = TableIndexDescriptor(
            name="unk_ti",
            definition=TableUnsupportedIndexDefinition(
                column="UNKNOWN",
                api_support=TableAPIIndexSupportDescriptor(
                    cql_definition="c", create_index=False, filter=True
                ),
            ),
            index_type=TableIndexType.UNKNOWN,
        )
        unk_ti_s = TableIndexDescriptor(
            name="unk_ti",
            definition=TableUnsupportedIndexDefinition(
                column="UNKNOWN",
                api_support=TableAPIIndexSupportDescriptor(
                    cql_definition="c", create_index=False, filter=True
                ),
            ),
            index_type="UNKNOWN",
        )
        assert unk_ti_0 == unk_ti_e
        assert unk_ti_0 == unk_ti_s



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from ..conftest import DefaultTable


@pytest.mark.skip("Feature not merged to main yet")
class TestTableGeneralBM25Sync:
    @pytest.mark.describe("test of table bm25 DML, sync")
    def test_table_bm25_dml_sync(
        self,
        sync_empty_table_textindex: DefaultTable,
    ) -> None:
        table = sync_empty_table_textindex
        table.insert_many(
            [
                {
                    "id": "tree_0",
                    "body": "A tree in the woods.",
                },
                {
                    "id": "tree_1",
                    "body": "The lone tree in the plain.",
                },
                {
                    "id": "fish",
                    "body": "A fish in the ocean.",
                },
            ]
        )

        # filtering:
        frows_f = table.find({"body": {"$match": "fish"}}).to_list()
        assert frows_f != []
        assert frows_f[0]["id"] == "fish"

        frow_f = table.find_one({"body": {"$match": "fish"}})
        assert frow_f is not None
        assert frow_f["id"] == "fish"

        frows_t = table.find({"body": {"$match": "tree"}}).to_list()
        assert {frow["id"] for frow in frows_t} == {"tree_0", "tree_1"}

        # no-match filtering:
        frow_n = table.find_one({"body": {"$match": "toadstool"}})
        assert frow_n is None

        frows_n = table.find({"body": {"$match": "toadstool"}}).to_list()
        assert frows_n == []

        # sorting:
        srows_f = table.find(sort={"body": "fish"}).to_list()
        assert srows_f != []
        assert srows_f[0]["id"] == "fish"

        srows_t = table.find(sort={"body": "tree"}).to_list()
        assert {srow["id"] for srow in srows_t[:2]} == {"tree_0", "tree_1"}

        srow_f = table.find_one(sort={"body": "fish"})
        assert srow_f is not None
        assert srow_f["id"] == "fish"

        # no-match sorting
        srows_n = table.find(sort={"body": "toadstool"}).to_list()
        assert srows_n == []

        srow_n = table.find_one(sort={"body": "toadstool"})
        assert srow_n is None
